# -*- coding: utf-8 -*-
"""
Last modified on Nov 13, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         Université Paris-Saclay, France

Contains the following classes for plotting purposes of program PyMaGra:

    - plot
    - newWindow

"""

import os
import numpy as np
from PyQt5.uic import loadUiType
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QVBoxLayout, QWidget
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure
from matplotlib.ticker import AutoMinorLocator
from matplotlib.path import Path as P
from matplotlib.gridspec import GridSpec
from matplotlib import tri
from ..inversion import mag_grav_utilities as utils

Ui_MainWindow, QMainWindow = loadUiType(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "Magnetics.ui"))


class mainWindow(QMainWindow, Ui_MainWindow):
    """
    Class for plotting in PyMaGra

    Depends on the following packages:

        - numpy
        - PyQt5
        - matplotlib

    Needed private files:

        - Magnetics.ui (user interface, produced with QtDesigner)

    Contains the following methods:

        - __init__
        - setClipPath
        - set_geography_flag
        - toggle_geography_flag
        - plot_geography
        - plot_lineaments
        - plot_north
        - plot_triang
        - plot_image
        - addMPL
        - rmMPL
        - plotFloating
        - get_mouse_click (with method onClick);
          Wait for mouse click return position and type
        - get_event (with method onClick);
          Wait for mouse click or keyboard press
        - setHelp : Set help text written at the base of the screen
        - followLine : Follow the cursor and plot a straight line

    """

    def __init__(self, main):
        super().__init__()
# Set up main window based on file Magnetics.ui created with QT Designer
        self.setupUi(self)
# create a first figure in central widget
        self.fig = Figure()
        self.fig.tight_layout()
        self.addMPL(self.fig)
        self.main = main
        self.grid_flag = True
        self.plotLin_flag = False
        self.geography_flag = False
        self.legend = []
        self.grad_data = False
        self.unit = "nT"
        self.bar_or = "vertical"
        self.nticks = 10
        self.axfont = 16
        self.cidmotion = None
        self.cidrelease = None
        self.dfac = 1.

    def setClipPath(self, x, y):
        """
        define clip path

        Parameters
        ----------
        x : numpy float array
            contains all x-coordinates of data points.
        y : numpy float array
            contains all y-coordinates of data points.

        Returns
        -------
        PATH matplotlib.path

        """
        clip_x = []
        clip_y = []
        if self.main.direction == 0:
            for xx in np.unique(x):
                yy = np.round(np.nanmax(y[x == xx]), 2)
                if len(clip_x) == 0:
                    yy0 = yy
                else:
                    if yy > yy0:
                        # clip_x.append(xx - 0.01)
                        xc = (xx+clip_x[-1])*0.5
                        clip_x.append(xc)
                        clip_y.append(yy0)
                        clip_x.append(xc)
                        clip_y.append(yy)
                    elif yy < yy0:
                        # clip_x.append(clip_x[-1] + 0.01)
                        # clip_y.append(yy0)
                        xc = (xx+clip_x[-1])*0.5
                        clip_x.append(xc)
                        clip_y.append(yy0)
                        clip_x.append(xc)
                        clip_y.append(yy)
                clip_x.append(xx)
                clip_y.append(yy)
                yy0 = yy
            il = 0
            for xx in np.unique(x)[::-1]:
                yy = np.round(np.nanmin(y[x == xx]), 2)
                if il == 0:
                    yy0 = yy
                    il = 1
                else:
                    if yy < yy0:
                        # clip_x.append(xx + 0.01)
                        # clip_y.append(yy0)
                        xc = (xx+clip_x[-1])*0.5
                        clip_x.append(xc)
                        clip_y.append(yy0)
                        clip_x.append(xc)
                        clip_y.append(yy)
                    elif yy > yy0:
                        # clip_x.append(clip_x[-1] - 0.01)
                        # clip_y.append(yy)
                        xc = (xx+clip_x[-1])*0.5
                        clip_x.append(xc)
                        clip_y.append(yy0)
                        clip_x.append(xc)
                        clip_y.append(yy)
                clip_x.append(xx)
                clip_y.append(yy)
                yy0 = yy
        else:
            for yy in np.unique(y):
                xx = np.round(np.nanmax(x[y == yy]), 2)
                if len(clip_x) == 0:
                    xx0 = xx
                else:
                    if xx > xx0:
                        # clip_y.append(yy - 0.01)
                        # clip_x.append(xx0)
                        yc = (yy+clip_y[-1])*0.5
                        clip_y.append(yc)
                        clip_x.append(xx0)
                        clip_y.append(yc)
                        clip_x.append(xx)
                    elif xx < xx0:
                        # clip_y.append(clip_y[-1] + 0.01)
                        # clip_x.append(xx)
                        yc = (yy+clip_y[-1])*0.5
                        clip_y.append(yc)
                        clip_x.append(xx0)
                        clip_y.append(yc)
                        clip_x.append(xx)
                clip_y.append(yy)
                clip_x.append(xx)
                xx0 = xx
            il = 0
            for yy in np.unique(y)[::-1]:
                xx = np.round(np.nanmin(x[y == yy]), 2)
                if il == 0:
                    xx0 = xx
                    il = 1
                else:
                    if xx < xx0:
                        # clip_y.append(yy + 0.01)
                        # clip_x.append(xx0)
                        yc = (yy+clip_y[-1])*0.5
                        clip_y.append(yc)
                        clip_x.append(xx0)
                        clip_y.append(yc)
                        clip_x.append(xx)
                    elif xx > xx0:
                        # clip_y.append(clip_y[-1] - 0.01)
                        # clip_x.append(xx)
                        yc = (yy+clip_y[-1])*0.5
                        clip_y.append(yc)
                        clip_x.append(xx0)
                        clip_y.append(yc)
                        clip_x.append(xx)
                clip_y.append(yy)
                clip_x.append(xx)
                xx0 = xx
        clip_x.append(clip_x[0])
        clip_y.append(clip_y[0])
        clip = np.zeros((len(clip_x), 2))
        clip[:, 0] = np.array(clip_x)
        clip[:, 1] = np.array(clip_y)
# Clip contains the coordinates of the clipping path.
# Codes will contain the way to connect clipping points (move to the first
# point of the path, draw lines to all other points and finally close the path)
        codes = []
        Path = P
        codes += [Path.MOVETO]
        codes += [Path.LINETO] * (len(clip_x) - 2)
        codes += [Path.CLOSEPOLY]
# set clipping
        return Path(clip, codes)

    def set_geography_flag(self, option):
        """
        Set flag to plot geography information or not to value "option".

        Parameters
        ----------
        option : bool
            If True, geography information is plotted.

        """
        self.geography_flag = option

    def toggle_geography_flag(self):
        """
        Inverse value of flag to plot geography information or not.

        No input or output parameters

        """
        self.geography_flag = not self.geography_flag

    def plot_geography(self, ax, dfac):
        """
        Plot geography information into axis "ax".

        Parameters
        ----------
        ax : Matplotlib.Axis object
            Axis where to plot the geography information
        dfac : float, optional; Default: 1.
            Factor by which to multiply fthe coordinates (usually conversion
            from m to km)

        Geography information is stored in class Main as dictionary with
        entries

        - "type" : may be "point" or "line"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line

        """
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
# Plot a point
        for key in self.main.geography.keys():
            x = np.array(self.main.geography[key]["x"])*dfac
            y = np.array(self.main.geography[key]["y"])*dfac
# If fpoint outside map don't plot
            if self.main.geography[key]["type"] == "POINT":
                if x < xmin or x > xmax or y < ymin or y > ymax:
                    continue
# Check whether a point has already been plotted. If not, add label to legend
                if "town" in self.legend:
                    ax.plot(x, y, "o", color="black")
                else:
                    self.legend.append("town")
                    ax.plot(x, y, "o", color="black", label="town/place")
                ax.text(x, y, "  " + self.main.geography[key]["name"],
                        ha="left", va="center", fontsize=12)
# Plot a line and check whether a line has already been plotted.
            else:
                if "geology" in self.legend:
                    ax.plot(x, y, "k")
                else:
                    self.legend.append("geology")
                    ax.plot(x, y, "k", label="geology")

    def plot_lineaments(self, ax, dfac=1.):
        """
        Plot lineaments picked from tilt maps into axis "ax".

        Parameters
        ----------
        ax : Matplotlib.Axis object
            axis where to plot the lineaments
        dfac : float, optional; Default: 1.
            Factor by which to multiply fthe coordinates (usually conversion
            from m to km)
        Lineament information is stored as dictionary with entries

        - "type" : may be "gravity" or "magnetic"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line

        """
        col = "k"
        self.lineaments = self.main.lineaments
        for key in self.lineaments.keys():
            x = np.array(self.lineaments[key]["x"])*dfac
            y = np.array(self.lineaments[key]["y"])*dfac
            if "gravity" in self.lineaments[key]["type"]:
                col = "w"
            elif "magnetic" in self.lineaments[key]["type"]:
                col = "k"
            if self.lineaments[key]["type"] in self.legend:
                ax.plot(x, y, col, ls="--", linewidth=3)
            else:
                self.legend.append(self.lineaments[key]["type"])
                ax.plot(x, y, col, label=self.lineaments[key]["type"], ls="--",
                        linewidth=3)

    def plot_north(self, ax, angle, pos):
        """
        Plot North arrow into map
        It is supposed that the map is drawn with set_aspect="equal"

        Parameters
        ----------
        ax : matplotlib axes object
            axis where to plot the North arrow.
        angle : float
            Angle in degrees from N to E (geological sense, not mathematical).
        pos : str
            string of two characters indicationg the position of the arrow.
            One character may be "t" or "b" for top or bottom of the plot
            The other letter may be "l" or "r" for left or right
            The order of the two characters as well as capitalization has no
            importance.

        Returns
        -------
        None.

        """
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        dx = (xmax - xmin) * 0.02
        dy = (ymax - ymin) * 0.05
        alpha = 90.0 - angle
        if alpha < -180:
            alpha += 360.0
        if alpha > 180:
            alpha -= 360.0
        c = np.cos(np.deg2rad(alpha))
        s = np.sin(np.deg2rad(alpha))
        dxa = dy * c
        dya = dy * s
        radius = np.sqrt(dxa**2 + dya**2)

        if "t" in pos.lower():
            y0 = ymax - (1.5 + s) * dy
        else:
            y0 = ymin + (1.5 - s) * dy
        if "r" in pos.lower():
            x0 = xmax - (2.5 + 2 * c) * dx
        else:
            x0 = xmin + (2.5 - 2 * c) * dx

        ax.arrow(x0, y0, dxa, dya, width=dya / 5.0, length_includes_head=True,
                 head_length=radius, head_width=radius / 1.5, shape="left",
                 color="k", edgecolor=None)
        ax.arrow(x0, y0, dxa, dya, width=dya / 5.0, length_includes_head=True,
                 head_length=radius, head_width=radius / 1.5, shape="right",
                 facecolor="w", edgecolor="k")
        ax.text(x0 + dxa * 0.5, y0 + dya * 0.5, "N", fontsize=20, c="r",
                horizontalalignment="center",
                verticalalignment="center_baseline",
                rotation=alpha - 90.0, rotation_mode="anchor")
        # ax.text(x0 + dxa * 1.6, y0 + dya * 1.75, "N",
        #         horizontalalignment="center",
        #         verticalalignment="center_baseline",
        #         rotation=alpha - 90.0, rotation_mode="anchor")

    def plot_sensor_triang(self, ax, x, y, s, vmin, vmax, percent, cmap,
                           xlabel, ylabel, title, cbar_title, dec=0.0,
                           dfac=1.):
        """
        Plot map of one sensor by triangulation

        Parameters
        ----------
        ax : Matplotlib axis
            Axis where to plot the map
        xx : Numpy 1D float array
            X-coordinates of data
        yy : Numpy 1D float array
            Y-coordinates of data
        s : Numpy 1D float array
            Values [nT or mGal]
        vmin : float
            Value for lower end of color scale [nT or mGal]
            If vmin==vmax, the color scale is calculated automatically based on
            quantiles (see "percent")
        vmax : float
            Value for upper end of color scale [nT or mGal]
        percent : float
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. This is only used if
            mincol==maxcol.
        cmap : Matplotlib color map
            Color map to be used
        xlabel : str
            Label for x-axis
        ylabel : str
            Label for y-axis
        title : str
            Title of map
        cbar_title : str
            Text for color bar
        dec : float (optional, default : 0.)
            Direction of the measurement lines with respect to magnetic North
            in degrees, measured from N to E

        Returns
        -------
        None.

        """
        self.legend = []
# Define clip paths for both sensors, searching for every line maximum and
#    minimum coordinates
        xmin = x.min()
        xmax = x.max()
        ymin = y.min()
        ymax = y.max()
        clip_path = self.setClipPath(x, y)
        rd = (int(np.ceil(-np.log10(np.nanquantile(s, 0.999) -
                                    np.nanquantile(s, 0.001)))) + 2)
# Define collor bar limits and levels
        if vmin == vmax:
            if percent > 0:
                max_col = np.round(np.nanquantile(s, 1 - percent), rd)
                min_col = np.round(np.nanquantile(s, percent), rd)
            else:
                max_col = np.round(np.nanmax(s), rd)
                min_col = np.round(np.nanmin(s), rd)
        else:
            max_col = np.round(vmax, rd)
            min_col = np.round(vmin, rd)
        if cmap == "special mag":
            cmap, norm = utils.mag_color_map(min_col, max_col, cif=2)
        else:
            norm = "linear"
        rd = int(np.ceil(-np.log10(max_col - min_col))) + 2
        nlev = 51
        ds = np.round((max_col - min_col) / nlev, rd)
        levels = np.arange(min_col, max_col + ds / 2, ds)
# Calculate triangulation
        triang = tri.Triangulation(x, y)
# Do triangulation plot
        gci = ax.tricontourf(triang, s, extend="both", vmin=min_col,
                             vmax=max_col, levels=levels, cmap=cmap, norm=norm)
        ax.set_clip_on(True)
        for collection in gci.collections:
            collection.set_clip_path(clip_path, transform=ax.transData)
        ax.set_title(title, fontsize=self.axfont+2)
        ax.set_xlabel(xlabel, fontsize=self.axfont)
        ax.set_ylabel(ylabel, fontsize=self.axfont)
        ax.tick_params(labelsize=self.axfont)
        if self.grid_flag:
            ax.grid(visible=True, which="both")
        if self.plotLin_flag:
            self.plot_lineaments(ax, dfac)
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        if self.bar_or == "vertical":
            cax = ax.inset_axes([1.05, 0.05, 0.02, 0.9],
                                transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.2, 0.9, 0.05],
                                transform=ax.transAxes)
        smin = np.round(np.nanmin(s), rd)
        smax = np.round(np.nanmax(s), rd)
        ssmin = levels[0]
        ssmax = levels[-1]
        ds = (ssmax - ssmin) / (self.nticks-1)
        ticks = np.round(np.arange(ssmin, ssmax+ds/2, ds), rd)
        ticks = list(ticks)
# Plot color bar
        cbar = plt.colorbar(gci, orientation=self.bar_or, cax=cax,
                            fraction=0.1, ticks=ticks, extend="both")
        if self.bar_or == "vertical":
            cbar.ax.set_ylabel(cbar_title, size=10)
            cbar.ax.text(0.0, -0.06, f"{smin}", verticalalignment="top",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
            cbar.ax.text(0.0, 1.06, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
        else:
            cbar.ax.set_xlabel(cbar_title, size=10)
            cbar.ax.text(0.0, 1.0, f"{smin}", verticalalignment="bottom",
                         horizontalalignment="right",
                         transform=cbar.ax.transAxes, fontsize=10)
            cbar.ax.text(1.0, 1.0, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
        ticklabs = cbar.ax.get_yticklabels()
        cbar.ax.set_yticklabels(ticklabs, fontsize=10)
        ticklabs = cbar.ax.get_xticklabels()
        cbar.ax.set_xticklabels(ticklabs, fontsize=10)
        ax.set_aspect("equal", adjustable="box")
        if self.geography_flag:
            self.plot_geography(ax, dfac)
        if self.plotLin_flag:
            ax.legend(bbox_to_anchor=(1, 1), loc="upper right", fontsize=10)
        self.plot_north(ax, -dec, "tl")

    def plot_triang(self, data, title="Measured magnetic data",
                    xlabel="Easting", ylabel="Northing", percent=0.01,
                    c="rainbow", mincol1=0.0, maxcol1=0.0, mincol2=0.0,
                    maxcol2=0.0, mincolg=0.0, maxcolg=0.0, grad_flag=False):
        """
        Plot maps of not interpolated data (stn format) of both sensors using
        tricontourf

        Parameters
        ----------
        data: dictionary with one entry per line, key = number of line.
              Each entry is itself a dictionary containing the following
              entries:

              - "s1": Numpy float array with data of sensor 1
              - "s2": Numpy float array with data of sensor 2
                If only data of one single sensor were measured, "s2" contains
                only one zero.
              - "x":  Numpy float array with E-W coordinates of data points
              - "y":  Numpy float array with N-S coordinates of data points
              - "grad_flag" bool. True if 2 sensors were used, False if only
                one sensor
              - "mask": bool, True if line should be plotted, False if excluded
                from plotting
        title : str, optional. Default is "Measured magnetic data"
            Plot title for both sensor polts (the text "sensor n" is
            automatically added, n being sensor numnber).
        xlabel : str, optional
            Text for x-axis. The default is "Easting".
        ylabel : str, optional
            Text for y-axis. The default is "Northing".
        percent : float, optional
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. The default is 0.01.
            This is only used if mincol==maxcol.
        mincol1 : float, optional
            value for lower end of color scale for sensor 1 [nT]
        maxcol1 : float, optional
            value for upper end of color scale for sensor 1 [nT]
            If mincol1==maxcol1, the color scale is calculated automatically
            based on quantiles (see "percent")
        mincol2 : float, optional
            value for lower end of color scale for sensor 2 [nT]
        maxcol2 : float, optional
            value for upper end of color scale for sensor 2 [nT]
            If mincol2==maxcol2, the color scale is calculated automatically
            based on quantiles (see "percent")
        mincolg : float, optional
            value for lower end of color scale for vertical gradient [nT/m]
        maxcolg : float, optional
            value for upper end of color scale for vertical gradient [nT/m]
            If mincolg==maxcolg, the color scale is calculated automatically
            based on quantiles (see "percent")
        c : str, optional
            Color map to be used. The default is "rainbow".
        grad_flag: bool, optional, default: False
            If True, 3 plots are shown on the screen: Sensor1, sensor2 and
            gradient. If False only the two sensors are plotted

        Returns
        -------
        fig : matplotlib.figure
            Figure used for plot.
        ax : matplotlib.axis
            List of axes used for plots. ax has two entrances, one for each
            sensor

        """
        x = []
        y = []
        v1 = []
        v2 = []
        self.grad_data = data.grad_data
        if data.data["type"] == "magnetic":
            self.unit = "nT"
        else:
            self.unit = "mGal"
        for key, val in data.data.items():
            try:
                _ = int(key)
                if not val["mask"]:
                    continue
                x += list(val["x"])
                y += list(val["y"])
                v1 += list(val["s1"])
                if self.grad_data:
                    v2 += list(val["s2"])
            except ValueError:
                break
        x = np.array(x)
        y = np.array(y)
        if y.max() > 100000.:
            x *= 0.001
            y *= 0.001
            dfac = 0.001
            dunit = "km"
        else:
            dfac = 1.
            dunit = "m"
        self.dfac = dfac
        if xlabel == "Easting":
            xlabel += f" [{dunit}]"
        if ylabel == "Northing":
            ylabel += f" [{dunit}]"
        v1 = np.array(v1)
        v2 = np.array(v2)
        ddx = x.max() - x.min()
        ddy = y.max() - y.min()

        ax = []
# Gradient is plotted
        ddx = data.xmax - data.xmin
        ddy = data.ymax - data.ymin
# Vertical layout
        if self.grad_data:
            if grad_flag:
                if ddx > 2.*ddy:
                    self.bar_or = "vertical"
                    self.nticks = 10
                    fig = plt.figure(figsize=(14, 12), layout="constrained")
                    self.gs = GridSpec(26, 10, fig)
                    ax.append(fig.add_subplot(self.gs[1:9, 1:]))
                    ax.append(fig.add_subplot(self.gs[9:17, 1:]))
                    ax.append(fig.add_subplot(self.gs[17:25, 1:]))
# Horizontal layout
                else:
                    fig = plt.figure(figsize=(26, 12), layout="constrained")
                    self.gs = GridSpec(10, 26, fig)
                    ax.append(fig.add_subplot(self.gs[1:, 1:9]))
                    ax.append(fig.add_subplot(self.gs[1:, 9:17]))
                    ax.append(fig.add_subplot(self.gs[1:, 17:25]))
                    self.bar_or = "horizontal"
                    self.nticks = 4
# Gradient is not plotted
# Vertical layout
            else:
                if ddx > 2.*ddy:
                    fig = plt.figure(figsize=(14, 12), layout="constrained")
                    self.gs = GridSpec(18, 10, fig)
                    ax.append(fig.add_subplot(self.gs[1:9, 1:]))
                    ax.append(fig.add_subplot(self.gs[9:17, 1:]))
                    self.bar_or = "vertical"
                    self.nticks = 10
# Horizontal layout
                else:
                    fig = plt.figure(figsize=(18, 12), layout="constrained")
                    self.gs = GridSpec(10, 18, fig)
                    ax.append(fig.add_subplot(self.gs[1:, 1:9]))
                    ax.append(fig.add_subplot(self.gs[1:, 9:17]))
                    self.bar_or = "horizontal"
                    self.nticks = 5
# Only sensor 1 is plotted
        else:
            fig = plt.figure(figsize=(12, 12), layout="constrained")
            self.gs = GridSpec(10, 10, fig)
            ax.append(fig.add_subplot(self.gs[1:, 1:9]))
            self.bar_or = "vertical"
            self.nticks = 10
        matplotlib.rcParams.update({"font.size": 14})
        self.axfont = 12
        plt.rc("xtick", labelsize=self.axfont)
        plt.rc("ytick", labelsize=self.axfont)

# Eliminate possible nans (coming from function clean_data) for both sensors
        x1 = np.copy(x)
        y1 = np.copy(y)
        s1 = np.copy(v1)
        if self.grad_data:
            x2 = np.copy(x)
            y2 = np.copy(y)
            s2 = np.copy(v2)
            x3 = np.copy(x)
            y3 = np.copy(y)
            s3 = s2 - s1

            index = np.isfinite(s1) & np.isfinite(s2)
            x3 = x3[index]
            y3 = y3[index]
            s3 = s3[index]
            x2 = np.delete(x2, np.isnan(s2))
            y2 = np.delete(y2, np.isnan(s2))
            s2 = np.delete(s2, np.isnan(s2))
        x1 = np.delete(x1, np.isnan(s1))
        y1 = np.delete(y1, np.isnan(s1))
        s1 = np.delete(s1, np.isnan(s1))

# Do plot for sensor 1
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 1;" + t[1]
            else:
                txt = title + " Sensor 1"
        else:
            txt = title
        self.plot_sensor_triang(ax[0], x1, y1, s1, mincol1, maxcol1, percent,
                                c, xlabel, ylabel, txt,
                                f"Field strength [{self.unit}]",
                                dec=data.line_declination, dfac=dfac)
# Do plot for sensor 2
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 2;" + t[1]
            else:
                txt = title + " Sensor 2"
            self.plot_sensor_triang(ax[1], x2, y2, s2, mincol2, maxcol2,
                                    percent, c, xlabel, ylabel, txt,
                                    f"Field strength [{self.unit}]",
                                    dec=data.line_declination, dfac=dfac)
# plot vertical gradient
            if grad_flag:
                if ";" in title[:-1]:
                    t = title.split(";")
                    txt = t[0] + " Gradient;" + t[1]
                else:
                    txt = title + " Gradient"
                self.plot_sensor_triang(ax[2], x3, y3, s3, mincolg, maxcolg,
                                        percent, c, xlabel, ylabel, txt,
                                        f"Field gradient [{self.unit}/m]",
                                        dec=data.line_declination, dfac=dfac)
# Erase actual plot in central widget and plot the new one
        self.rmMPL()
        self.addMPL(fig)
        return fig, ax

    def plot_sensor_image(self, ax, x, y, s, vmin, vmax, percent, cmap,
                          xlabel, ylabel, title, cbar_title, dec=0.0, dfac=1.):
        """
        Plot map of one sensor by triangulation

        Parameters
        ----------
        ax : Matplotlib axis
            Axis where to plot the map
        x : Numpy 1D float array
            X-coordinates of data columns
        y : Numpy 1D float array
            Y-coordinates of data rows
        s : Numpy 2D float array
            Values [nT or mGal] on a regular grid
        vmin : float
            Value for lower end of color scale [nT or mGal]
            If vmin==vmax, the color scale is calculated automatically based on
            quantiles (see "percent")
        vmax : float
            Value for upper end of color scale [nT or mGal]
        percent : float
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. This is only used if
            mincol==maxcol.
        cmap : Matplotlib color map
            Color map to be used
        xlabel : str
            Label for x-axis
        ylabel : str
            Label for y-axis
        title : str
            Title of map
        cbar_title : str
            Text for color bar
        dec : float (optional, default : 0.)
            Direction of the measurement lines with respect to magnetic North
            in degrees, measured from N to E

        Returns
        -------
        None.

        """
        self.legend = []
        matplotlib.rcParams.update({"font.size": 16})
        axfont = 12
        matplotlib.rcParams.update({"font.size": 16})
        plt.rc("xtick", labelsize=axfont)
        plt.rc("ytick", labelsize=axfont)
# Calculate color scale
        if vmin == vmax:
            if percent > 0:
                max_col = np.nanquantile(s, 1 - percent)
                min_col = np.nanquantile(s, percent)
            else:
                max_col = np.nanmax(s)
                min_col = np.nanmin(s)
        else:
            max_col = vmax
            min_col = vmin
        rd = int(np.ceil(-np.log10(max_col - min_col))) + 2
        xmin = np.nanmin(x)
        xmax = np.nanmax(x)
        ymin = np.nanmin(y)
        ymax = np.nanmax(y)
        dx2 = (x[1] - x[0]) / 2
        dy2 = (y[1] - y[0]) / 2
        if cmap == "special mag":
            cmap, norm = utils.mag_color_map(min_col, max_col, cif=2)
            im1 = ax.imshow(np.flip(s, axis=0), cmap=cmap, norm=norm,
                            extent=[xmin-dx2, xmax+dx2, ymin-dy2, ymax+dy2],
                            aspect="equal")
        else:
            norm = "linear"
        im1 = ax.imshow(np.flip(s, axis=0), cmap=cmap, aspect="equal",
                        extent=[xmin-dx2, xmax+dx2, ymin-dy2, ymax+dy2],
                        vmin=min_col, vmax=max_col)
        if self.grad_data:
            ax.set_title(title, fontsize=14)
        else:
            ax.set_title(title, fontsize=14)
        ax.set_xlabel(xlabel, fontsize=axfont)
        ax.set_ylabel(ylabel, fontsize=axfont)
        ax.tick_params(labelsize=axfont)
        if self.grid_flag:
            ax.grid(visible=True, which="both")
        if self.plotLin_flag:
            self.plot_lineaments(ax, dfac)
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        if self.bar_or == "vertical":
            cax = ax.inset_axes([1.05, 0.05, 0.02, 0.9],
                                transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.2, 0.9, 0.05],
                                transform=ax.transAxes)
        smin = np.round(np.nanmin(s), rd)
        smax = np.round(np.nanmax(s), rd)
        ssmin = min_col
        ssmax = max_col
        ds = (ssmax - ssmin) / self.nticks
        ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
        ticks = list(ticks)
        labels = []
        for t in ticks:
            labels.append(f"{t}")
        cbar = plt.colorbar(im1, orientation=self.bar_or, cax=cax,
                            fraction=0.1, ticks=ticks, extend="both")
        if self.bar_or == "vertical":
            cbar.ax.set_ylabel(cbar_title, size=10)
            cbar.ax.text(0.0, -0.06, f"{smin}", verticalalignment="top",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
            cbar.ax.text(0.0, 1.06, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
        else:
            cbar.ax.set_xlabel(cbar_title, size=10)
            cbar.ax.text(0.0, 1.0, f"{smin}", verticalalignment="bottom",
                         horizontalalignment="right",
                         transform=cbar.ax.transAxes, fontsize=10)
            cbar.ax.text(1.0, 1.0, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=10)
        ticklabs = cbar.ax.get_yticklabels()
        cbar.ax.set_yticklabels(ticklabs, fontsize=10)
        ticklabs = cbar.ax.get_xticklabels()
        cbar.ax.set_xticklabels(ticklabs, fontsize=10)
        if self.geography_flag:
            self.plot_geography(ax, dfac)
        #        ax[0].set_aspect('equal', adjustable='box', anchor=anchor)
        ax.set_aspect("equal", adjustable="box")
        if self.plotLin_flag:
            ax.legend(bbox_to_anchor=(1, 1), loc="upper right", fontsize=10)
        self.plot_north(ax, -dec, "tl")

    def plot_image(self, data, title="Measured magnetic data",
                   xlabel="Easting", ylabel="Northing", percent=0.01,
                   c="rainbow", mincol1=0, maxcol1=0, mincol2=0.0, maxcol2=0.0,
                   mincolg=0.0, maxcolg=0.0, grad_flag=False, dec=0.0):
        """
        Plot maps of interpolated data of both sensors using imshow

        Parameters
        ----------
        data : DataContainer object
            Contains data to be plotted
        title : str, optional. Default is "Measured magnetic data"
            Plot title for both sensor polts (the text "sensor n" is
            automatically added, n being sensor numnber).
        xlabel : str, optional
            Text for x-axis. The default is "Easting".
        ylabel : str, optional
            Text for y-axis. The default is "Northing".
        percent : float, optional
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. The default is 0.01.
        mincol1 : float, optional
            value for lower end of color scale for sensor 1 [nT]
        maxcol1 : float, optional
            value for upper end of color scale for sensor 1 [nT]
            If mincol1==maxcol1, the color scale is calculated automatically
            based on quantiles (see "percent")
        mincol2 : float, optional
            value for lower end of color scale for sensor 2 [nT]
        maxcol2 : float, optional
            value for upper end of color scale for sensor 2 [nT]
            If mincol2==maxcol2, the color scale is calculated automatically
            based on quantiles (see "percent")
        mincolg : float, optional
            value for lower end of color scale for vertical gradient [nT/m]
        maxcolg : float, optional
            value for upper end of color scale for vertical gradient [nT/m]
            If mincolg==maxcolg, the color scale is calculated automatically
            based on quantiles (see "percent")
        c : str, optional
            Color map to be used. The default is "rainbow".
        grad_flag: bool, optional, default: False
            If True, 3 plots are shown on the screen: Sensor1, sensor2 and
            gradient. If False only the two sensors are plotted.
        dec : float
            Y direction with respect to geographic north (measured from N to E)

        Returns
        -------
        fig : matplotlib.figure
            Figure used for plot.
        ax : matplotlib.axis
            List of axes used for plots. ax has two entrances, one for each
            sensor

        """
        self.legend = []
        ax = []
        ddx = data.x_inter.max() - data.x_inter.min()
        ddy = data.y_inter.max() - data.y_inter.min()
        if data.y_inter.max() > 100000.:
            x = np.copy(data.x_inter)*0.001
            y = np.copy(data.y_inter)*0.001
            dfac = 0.001
            dunit = "km"
        else:
            dfac = 1.
            x = np.copy(data.x_inter)
            y = np.copy(data.y_inter)
            dunit = "m"
        self.dfac = dfac
        if xlabel == "Easting":
            xlabel += f" [{dunit}]"
        if ylabel == "Northing":
            ylabel += f" [{dunit}]"
# Gradientplotted
# Vertical layout
        if self.grad_data:
            if grad_flag:
                if ddx > 2.*ddy:
                    fig = plt.figure(figsize=(14, 12), layout="constrained")
                    self.gs = GridSpec(26, 10, fig)
                    ax.append(fig.add_subplot(self.gs[1:9, 1:]))
                    ax.append(fig.add_subplot(self.gs[9:17, 1:]))
                    ax.append(fig.add_subplot(self.gs[17:25, 1:]))
                    self.bar_or = "vertical"
#                    anchor = 'E'
                    self.nticks = 10
# Horizontal layout
                else:
                    fig = plt.figure(figsize=(26, 12), layout="constrained")
                    self.gs = GridSpec(10, 26, fig)
                    ax.append(fig.add_subplot(self.gs[1:, 1:9]))
                    ax.append(fig.add_subplot(self.gs[1:, 9:17]))
                    ax.append(fig.add_subplot(self.gs[1:, 17:25]))
                    self.bar_or = "horizontal"
#                    anchor = 'S'
                    self.nticks = 4
# Gradient not plotted
# Vertical layout
            else:
                if ddx > 1.5*ddy:
                    fig = plt.figure(figsize=(14, 12), layout="constrained")
                    self.gs = GridSpec(18, 10, fig)
                    ax.append(fig.add_subplot(self.gs[1:9, 1:]))
                    ax.append(fig.add_subplot(self.gs[9:17, 1:]))
                    self.bar_or = "vertical"
#                    anchor = 'E'
                    self.nticks = 10
# Horizontal layout
                else:
                    fig = plt.figure(figsize=(18, 12), layout="constrained")
                    self.gs = GridSpec(10, 18, fig)
                    ax.append(fig.add_subplot(self.gs[1:, 1:9]))
                    ax.append(fig.add_subplot(self.gs[1:, 9:17]))
                    self.bar_or = "horizontal"
#                    anchor = 'S'
                    self.nticks = 5
        else:
            fig = plt.figure(figsize=(12, 12), layout="constrained")
            self.gs = GridSpec(10, 10, fig)
            ax.append(fig.add_subplot(self.gs[1:, 1:9]))
            self.bar_or = "vertical"
#            anchor = 'E'
            self.nticks = 10

# Do plot for sensor 1
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 1;" + t[1]
            else:
                txt = title + " Sensor 1"
        else:
            txt = title
        self.plot_sensor_image(ax[0], x, y,
                               data.sensor1_inter, mincol1, maxcol1,
                               percent, c, xlabel, ylabel, txt,
                               f"Field strength [{self.unit}]", dec=dec,
                               dfac=dfac)

# Do plot for sensor 2
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 2;" + t[1]
            else:
                txt = title + " Sensor 2"
            self.plot_sensor_image(ax[1], x, y,
                                   data.sensor2_inter, mincol2, maxcol2,
                                   percent, c, xlabel, ylabel, txt,
                                   f"Field strength [{self.unit}]", dec=dec,
                                   dfac=dfac)
# Do plot for gradient
            if grad_flag:
                if ";" in title[:-1]:
                    t = title.split(";")
                    txt = t[0] + " Gradient;" + t[1]
                else:
                    txt = title + " Gradient"
                self.plot_sensor_image(ax[2], x, y,
                                       data.grad_inter, mincolg, maxcolg,
                                       percent, c, xlabel, ylabel, txt,
                                       f"Gradient [{self.unit}/m]", dec=dec,
                                       dfac=dfac)
# Erase actual plot in central widget and plot the new one
        self.rmMPL()
        self.addMPL(fig)
        return fig, ax

    def addMPL(self, fig):
        """
        Add widget to actual Figure

        Parameters
        ----------
        fig : matplotlib.figure.Figure object
            Actually used Figure

        """
        self.canvas = FigureCanvas(fig)
        self.mplvl.addWidget(self.canvas)
        self.canvas.draw()
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.mplvl.addWidget(self.toolbar)

    def rmMPL(self):
        """
        Remove widget from actual Figure
        """
        self.mplvl.removeWidget(self.canvas)
        self.canvas.close()
        self.mplvl.removeWidget(self.toolbar)
        self.toolbar.close()

#     def plotFloating(self, data, x, y, wtitle="", sizeh=800, sizev=500,
#                      mincol=0, maxcol=0, percent=0, c="rainbow", ptitle="",
#                      xlabel="", ylabel="", clabel=""):
#         """
#         Plot one or several 2D arrays into floating window

#         Parameters
#         ----------
#         data : numpy array
#             Data to be plotted.
#             The array my be 2D or 3D. If 3D, it is supposed that several
#             figures should be plotted into the same window in different frames.
#             If more than one plot is to be done, the thrid dimension contains
#             the different data sets. Every data set must be 2D. Several data
#             sets may be concatenated into data with the following commands:

#             - data = data1.reshape(nr,nc,1)
#             - data = np.concatenate((data,data2.reshape(nr,nc,1)),axis=2)

#             data1 and data2 are arrays with shape (nr,nc) defined on regular
#             grids.

#         x : numpy 1D float array
#             Positions of the columns
#         y : numpy 1D float array
#             Positions of the rows
#         wtitle : str optional, default: empty string
#             Title of floating window
#         sizeh : float optional,default: 800
#             horizontal size of floating window.
#         sizev : float optional, default: 500
#             vertical size of floating window
#         mincol : float optional, default: 0
#             Minimum value of color scale
#         maxcol : float optional,  default: 0
#             Maximum value of color scale.
#             If mincol == maxcol,  and percent == 0, the limits of the color
#             scale are minimum and maximum values of the data.
#         percent : float optional:
#             If >0, mincol and maxcol are calculated as quantiles
#             (percent, 1-percent).
#         c: str optional, default: "rainbow"
#             color scale to be used.
#         ptitle : str, optional, default: empty string
#             May be a single string or a list of strings, where the length of
#             the list must correspond to the length of the third dimension of
#             data.
#         xlabel : str, optional
#             Similar as ptitle for lables of horizontal axis
#         ylabel : str, optional
#             Similar as ptitle for lables of vertical axis
#         clabel : str, optional
#             Similar as ptitle for lables of color bars

#         Returns
#         -------
#         fig_float,ax_float: Matplot values of figure and axes

#         """
# # Only 2D arrays may be plotted, test if 1D data are passed to the routine
#         if data.ndim < 2:
#             _ = QtWidgets.QMessageBox.warning(
#                 None, "Warning",
#                 "Function plotFloating is not prepared for 1D plots",
#                 QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
#             return False
# # Create figure
#         fig_float = newWindow(wtitle, sizeh, sizev)
#         fig_float.fig.tight_layout()
# # If 2D array is passed create single axis
#         if data.ndim == 2:
#             ax_float = fig_float.fig.subplots(1, 1)
#             data1 = np.copy(data)
#             bar_or = "vertical"
# #            anchor = 'E'
#             nticks = 10
# # If 3D array is passed create 2 or 3 axis depending on the shape of data
#         else:
#             ddx = x.max() - x.min()
#             ddy = y.max() - y.min()
#             data1 = data[:, :, 0]
# # If horizontal extension is > 1.5x vertical one, plot axes vertically one
# #    above the next. If not plot axis in horizontal direction
#             if data.shape[2] == 3:
#                 facx = sizeh / (3 * ddx)
#                 facy = sizev / (3 * ddy)
#                 if facx < facy:
#                     ax_float = fig_float.fig.subplots(3, 1)
#                     bar_or = "vertical"
# #                    anchor = 'E'
#                     nticks = 10
#                 else:
#                     ax_float = fig_float.fig.subplots(1, 3)
#                     bar_or = "horizontal"
# #                    anchor = 'S'
#                     nticks = 6
#             else:
#                 facx = sizeh / (2 * ddx)
#                 facy = sizev / (2 * ddy)
#                 if facx < facy:
#                     ax_float = fig_float.fig.subplots(2, 1)
#                     bar_or = "vertical"
# #                    anchor = 'E'
#                     nticks = 10
#                 else:
#                     ax_float = fig_float.fig.subplots(1, 2)
#                     bar_or = "horizontal"
# #                    anchor = 'S'
#                     nticks = 6
# # Calculate color scale for both sensors independently
#         rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
#         if mincol == maxcol:
#             if percent > 0:
#                 max_col = np.round(np.nanquantile(data1, 1 - percent), rd)
#                 min_col = np.round(np.nanquantile(data1, percent), rd)
#             else:
#                 max_col = np.nanmax(data1)
#                 min_col = np.nanmin(data1)
#         else:
#             max_col = maxcol
#             min_col = mincol
# # Plot first into axis ax_float[0]
#         try:
#             ax = ax_float[0]
#             pt = ptitle[0]
#             xl = xlabel[0]
#             yl = ylabel[0]
#             cl = clabel[0]
# # If only one subplot has been created
#         except TypeError:
#             ax = ax_float
#             pt = ptitle
#             xl = xlabel
#             yl = ylabel
#             cl = clabel
#         dx2 = (x[1] - x[0]) / 2
#         dy2 = (y[1] - y[0]) / 2
#         im1 = ax.imshow(np.flip(data1, axis=0), cmap=c, aspect="equal",
#                         extent=[np.min(x)-dx2, np.max(x)+dx2, np.min(y)-dy2,
#                                 np.max(y) + dy2], vmin=min_col, vmax=max_col)
#         ax.set_title(pt)
#         ax.set_xlabel(xl)
#         ax.set_ylabel(yl)
#         if bar_or == "vertical":
#             cax = ax.inset_axes([1.015, 0.05, 0.015, 0.9],
#                                 transform=ax.transAxes)
#         else:
#             cax = ax.inset_axes([0.05, -0.15, 0.9, 0.05],
#                                 transform=ax.transAxes)
#         smin = np.round(np.nanmin(data1), rd)
#         smax = np.round(np.nanmax(data1), rd)
#         ssmin = min_col
#         ssmax = max_col
#         ds = (ssmax - ssmin) / nticks
#         ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
#         ticks = list(ticks)
# #        cbar = plt.colorbar(
#         cbar = fig_float.fig.colorbar(
#             im1, orientation=bar_or, ax=ax, cax=cax, fraction=0.1,
#             extend="both", ticks=ticks)
#         if bar_or == "vertical":
#             cbar.ax.set_ylabel(cl)
#             cbar.ax.text(0.0, -0.075, f"{smin}", verticalalignment="top",
#                          horizontalalignment="left",
#                          transform=cbar.ax.transAxes)
#             cbar.ax.text(0.0, 1.075, f"{smax}", verticalalignment="bottom",
#                          horizontalalignment="left",
#                          transform=cbar.ax.transAxes)
#         else:
#             cbar.ax.set_xlabel(cl)
#             cbar.ax.text(0.0, 1.0, f"{smin}", verticalalignment="bottom",
#                          horizontalalignment="right",
#                          transform=cbar.ax.transAxes,)
#             cbar.ax.text(1.0, 1.0, f"{smax}", verticalalignment="bottom",
#                          horizontalalignment="left",
#                          transform=cbar.ax.transAxes)
#         if self.grid_flag:
#             ax.grid(visible=True)
#         #        ax.set_aspect('equal', adjustable='box', anchor=anchor)
#         ax.set_aspect("equal", adjustable="box")
# # If more than one map has to be plotted, do this now
#         if data.ndim > 2:
#             if data.shape[2] > 1:
#                 data1 = data[:, :, 1]
#                 rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
#                 if mincol == maxcol:
#                     if percent > 0:
#                         max_col = np.round(np.nanquantile(data1, 1 - percent),
#                                            rd)
#                         min_col = np.round(np.nanquantile(data1, percent), rd)
#                     else:
#                         max_col = np.nanmax(data1)
#                         min_col = np.nanmin(data1)
#                 else:
#                     max_col = maxcol
#                     min_col = mincol
#                 im1 = ax_float[1].imshow(np.flip(data1, axis=0), cmap=c,
#                                          aspect="equal",
#                                          extent=[np.min(x)-dx2, np.max(x)+dx2,
#                                                  np.min(y)-dy2, np.max(y)+dy2],
#                                          vmin=min_col, vmax=max_col)
#                 ax_float[1].set_title(ptitle[1])
#                 ax_float[1].set_xlabel(xlabel[1])
#                 ax_float[1].set_ylabel(ylabel[1])
#                 if self.grid_flag:
#                     ax_float[1].grid(visible=True)
#                 if bar_or == "vertical":
#                     cax = ax_float[1].inset_axes(
#                         [1.015, 0.05, 0.015, 0.9],
#                         transform=ax_float[1].transAxes)
#                 else:
#                     cax = ax_float[1].inset_axes(
#                         [0.05, -0.15, 0.9, 0.05],
#                         transform=ax_float[1].transAxes)
#                 smin = np.round(np.nanmin(data1), rd)
#                 smax = np.round(np.nanmax(data1), rd)
#                 ssmin = min_col
#                 ssmax = max_col
#                 ds = (ssmax - ssmin) / nticks
#                 ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
#                 ticks = list(ticks)
# #                cbar = plt.colorbar(
#                 cbar = fig_float.fig.colorbar(im1, orientation=bar_or, ax=ax,
#                                               cax=cax, fraction=0.1,
#                                               extend="both", ticks=ticks)
#                 if bar_or == "vertical":
#                     cbar.ax.set_ylabel(clabel[1])
#                     cbar.ax.text(0.0, -0.075, f"{smin}",
#                                  verticalalignment="top",
#                                  horizontalalignment="left",
#                                  transform=cbar.ax.transAxes)
#                     cbar.ax.text(0.0, 1.075, f"{smax}",
#                                  verticalalignment="bottom",
#                                  horizontalalignment="left",
#                                  transform=cbar.ax.transAxes)
#                 else:
#                     cbar.ax.set_xlabel(clabel[1])
#                     cbar.ax.text(0.0, 1.0, f"{smin}",
#                                  verticalalignment="bottom",
#                                  horizontalalignment="right",
#                                  transform=cbar.ax.transAxes)
#                     cbar.ax.text(1.0, 1.0, f"{smax}",
#                                  verticalalignment="bottom",
#                                  horizontalalignment="left",
#                                  transform=cbar.ax.transAxes)
# #                ax_float[1].set_aspect('equal', adjustable='box',
# #                                       anchor=anchor)
#                 ax_float[1].set_aspect("equal", adjustable="box")
#                 if data.shape[2] > 2:
#                     data1 = data[:, :, 2]
#                     rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
#                     if mincol == maxcol:
#                         if percent > 0:
#                             max_col = np.round(
#                                 np.nanquantile(data1, 1 - percent), rd)
#                             min_col = np.round(
#                                 np.nanquantile(data1, percent), rd)
#                         else:
#                             max_col = np.nanmax(data1)
#                             min_col = np.nanmin(data1)
#                     else:
#                         max_col = maxcol
#                         min_col = mincol
#                     im1 = ax_float[2].imshow(np.flip(
#                         data1, axis=0), cmap=c, aspect="equal",
#                         extent=[np.min(x)-dx2, np.max(x)+dx2,
#                                 np.min(y)-dy2, np.max(y)+dy2],
#                         vmin=min_col, vmax=max_col)
#                     ax_float[2].set_title(ptitle[2])
#                     ax_float[2].set_xlabel(xlabel[2])
#                     ax_float[2].set_ylabel(ylabel[2])
#                     if self.grid_flag:
#                         ax_float[2].grid(visible=True)
#                     if bar_or == "vertical":
#                         cax = ax_float[2].inset_axes(
#                             [1.015, 0.05, 0.015, 0.9],
#                             transform=ax_float[2].transAxes)
#                     else:
#                         cax = ax_float[2].inset_axes(
#                             [0.05, -0.15, 0.9, 0.05],
#                             transform=ax_float[2].transAxes)
#                     smin = np.round(np.nanmin(data1), rd)
#                     smax = np.round(np.nanmax(data1), rd)
#                     ssmin = min_col
#                     ssmax = max_col
#                     ds = (ssmax - ssmin) / nticks
#                     ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
#                     ticks = list(ticks)
# #                    cbar = plt.colorbar(
#                     cbar = fig_float.fig.colorbar(im1, orientation=bar_or,
#                                                   ax=ax, cax=cax, fraction=0.1,
#                                                   extend="both", ticks=ticks)
#                     if bar_or == "vertical":
#                         cbar.ax.set_ylabel(clabel[2])
#                         cbar.ax.text(0.0, -0.075, f"{smin}",
#                                      verticalalignment="top",
#                                      horizontalalignment="left",
#                                      transform=cbar.ax.transAxes)
#                         cbar.ax.text(0.0, 1.075, f"{smax}",
#                                      verticalalignment="bottom",
#                                      horizontalalignment="left",
#                                      transform=cbar.ax.transAxes)
#                     else:
#                         cbar.ax.set_xlabel(clabel[2])
#                         cbar.ax.text(0.0, 1.0, f"{smin}",
#                                      verticalalignment="bottom",
#                                      horizontalalignment="right",
#                                      transform=cbar.ax.transAxes)
#                         cbar.ax.text(1.0, 1.0, f"{smax}",
#                                      verticalalignment="bottom",
#                                      horizontalalignment="left",
#                                      transform=cbar.ax.transAxes)
#                     #       ax_float[2].set_aspect('equal', adjustable='box',
#                     #                              anchor=anchor)
#                     ax_float[2].set_aspect("equal", adjustable="box")
#         fig_float.show()
#         return fig_float, ax_float

    def get_mouse_click(self, fig):
        """
        wait for a mouse click within figure "fig".

        Parameters
        ----------
        fig : matplotlib figure
            for main window, it must be self.w.mplvl
            for floating window, it is the name that has been given to the
            window at creation time

        Returns
        -------
        event.
        Mouse position, button pressed and the general event parameters are
        accessible through event.xdata, event.ydata, event.button and
        others like event.x or event.y

        """
        self.wait = True
        self.click = False
        self.press = False

        def onClick(event):
            self.wait = False
            self.click = True
            self.event = event
            self.event.xdata /= self.dfac
            self.event.ydata /= self.dfac
            self.x_mouse = event.xdata
            self.y_mouse = event.ydata
            self.mouse = event.button
            self.xwin = event.x
            self.ywin = event.y

        fig.canvas.mpl_connect("button_press_event", onClick)
        while self.wait:
            QtCore.QCoreApplication.processEvents()
        return self.event

    def get_event(self, fig):
        """
        wait for a mouse click of keyboard press within figure "fig".

        Parameters
        ----------
        fig : matplotlib figure
            for main window, it must be self.w.mplvl
            for floating window, it is the name that has been given to the
            window at creation time

        Returns
        -------
        event.
        Mouse position, button pressed and the general event parameters are
        accessible through event.xdata, event.ydata, event.button and
        others like event.x or event.y

        """
        self.wait = True

        def onClick(event):
            self.wait = False
            self.event = event
            self.event.xdata /= self.dfac
            self.event.ydata /= self.dfac

        fig.canvas.mpl_connect("button_press_event", onClick)
        fig.canvas.mpl_connect("key_press_event", onClick)
        while self.wait:
            QtCore.QCoreApplication.processEvents()
        return self.event

    def setHelp(self, text):
        """
        Set help text at bottom of screen.

        Parameters
        ----------
        text : str
            Text to be printed (defined in __init__)
            Text is written in QLabel widget. In order to keep the widget at
            the bottom of the screen, the existing one is first erased and
            then reopened. This call has to be done after any modification of
            the graphics window.

        fig : Matplotlib Figure
            Figure where to write the help text.

        Returns
        -------
        None
        """
        try:
            self.help.close()
        except (AttributeError, RuntimeError):
            pass
        self.help = QtWidgets.QLabel(self)
        self.help.setMaximumHeight(20)
        self.mplvl.addWidget(self.help)
        self.help.show()
        self.help.setText(text)



# -*- coding: utf-8 -*-
"""
Last modified on Dec 02, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         Université Paris-Saclay, France

Contains the following utility functions for program PyMaGra:

    - extract : Mark data segments for treatment as function of odd or even
        line number
    - data_flatten : Copy dictionary data into Numpy 1D array
    - clean_data : Mute (set to nan) data values outside a user-defined range
    - julian2date : Transform julian day number to calender date
    - date2julian : Transform calender date to julian day number
    - time2data : Convert seconds from start of year into julian day and time
    - diurnal_variation : Estimate diurnal variations from line medians
    - diurnal_correction : Correct diurnal variations of magnetic data
    - interpol_line : Interpolate data along a line onto a regular grid
    - interpol_2D : Interpolate data onto a regular 2D grid
    - extrapolate : Extrapolate data to fille a full rectangle
    - justify_lines_median : Eliminate directional effect by median adjustment
    - gauss_transform : Calculate Gaussian statistics parameters
    - justify_lines_gaussian : Eliminate directional effect by adjusting
        Gaussian statistics
    - next_minute : Calculate next full minute after a given time
    - add_time : Add a time to a given one (in hour, minute, second)
    - fit2lines : Fit two straight lines to a data series
    - min_max : Calculate positions of local maxima and minima in a data series
    - min_max2D : Calculate position of local maxima and minima on a 2D grid
    - get_date_string : Transform actual date and time into a string
    - file_name : Create a file name containing the actual date and time

"""

import os
from copy import deepcopy
from datetime import datetime
import numpy as np
from scipy import interpolate
from sklearn.preprocessing import QuantileTransformer
from sklearn.linear_model import LinearRegression as LR
from PyQt5 import QtWidgets


def extract(data, choice):
    """
    Mark data segments for treatment as function of odd or even line number

    Parameters
    ----------
    choice : tuple of str
        may be:

        - "all" (choose all lines)
        - "odd" (choose odd lines, natural counting)
        - "even" (choose even line numbers, natural counting)
        - "N", "S", "W" or "E"

    """
    if choice == "all":
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            val["mask"] = True
        data.line_choice = "all"
    else:
        data.line_choice = choice
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            if choice == "odd":
                if key % 2 == 1:
                    val["mask"] = True
                else:
                    val["mask"] = False
            elif choice == "even":
                if key % 2 == 0:
                    val["mask"] = True
                else:
                    val["mask"] = False


def data_flatten(data):
    """
    Data that are stored in dictionary data with one entrance per line are
    concatenated into one 1D numpy array

    Parameters
    ----------
    data : dictionary with keys equal to line number
        contains for every line itself a dictionary with the following
        keys:

        - "s1" : Data of sensor 1
        - "s2" : Data of sensor 2
        - "x", "y", "z": Coordinates of data

    Returns
    -------
    s1 : Numpy 1D float array
         Concatenated data of sensor 1
    s2 : Numpy 1D float array
         Concatenated data of sensor 2
    x  : Numpy 1D float array
         Concatenated x-coordinates of all data
    y  : Numpy 1D float array
         Concatenated y-coordinates of all data
    z  : Numpy 1D float array
         Concatenated z-coordinates of all data
    """
    s1 = []
    s2 = []
    x = []
    y = []
    z = []
    for key, val in data.items():
        if isinstance(key, (str)):
            break
        s1 += list(val["s1"])
        s2 += list(val["s2"])
        x += list(val["x"])
        y += list(val["y"])
        z += list(val["z"])
    return s1, s2, x, y, z


def clean_data(data, min_fix=None, max_fix=None, percent_down=None,
               percent_up=None):
    """
    Set data to np.nan under certain conditions which may be:

    Parameters
    ----------
    data : object of class data
        Contains data to be cleaned (see io.get_line).
    min_fix : float
        All data below this value are set to nan.
    max_fix : float
        All data above this value are set to nan.
    percent_down : float
        The lowermost percentile values are set to nan
        A value of 0.01 means that all values lower than the
        1% quantile are set to None.
    percent_up : float
        The uppermost percentile values are set to nan
        A value of 0.01 means that all values higher than the
        99% quantile are set to None.

    Returns
    -------
    data : dictionary
        Same structure as input data, but with with data outside defined
        limits set to np.nan
    """
    print("\n")
    grad_data = data.grad_data
    # s1, s2, _, _, _ = data_flatten(data)
    s1 = data.sensor1
    s2 = data.sensor2
    # s1 = data
    if min_fix:
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            v = np.copy(val["s1"])
            v[v < min_fix] = np.nan
            val["s1"] = np.copy(v)
            if grad_data:
                v = np.copy(val["s2"])
                v[v < min_fix] = np.nan
                val["s2"] = np.copy(v)
        print(f"Clip below {np.round(min_fix,1)}")
    if max_fix:
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            v = np.copy(val["s1"])
            v[v > max_fix] = np.nan
            val["s1"] = np.copy(v)
            if grad_data:
                v = np.copy(val["s2"])
                v[v > max_fix] = np.nan
                val["s2"] = np.copy(v)
        print(f"Clip above {np.round(max_fix,1)}")
    if percent_down:
        vmin1 = np.nanquantile(s1, percent_down)
        if grad_data:
            vmin2 = np.nanquantile(s2, percent_down)
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            v = np.copy(val["s1"])
            v[v < vmin1] = np.nan
            val["s1"] = np.copy(v)
            if grad_data:
                v = np.copy(val["s2"])
                v[v < vmin2] = np.nan
                val["s2"] = np.copy(v)
        if grad_data:
            print(
                f"Clip below {np.round(vmin1,1)} for sensor 1 and "
                + f"{np.round(vmin2,1)} for sensor 2")
        else:
            print(f"Clip below {np.round(vmin1,1)}")
    if percent_up:
        vmax1 = np.nanquantile(s1, 1.0 - percent_up)
        if grad_data:
            vmax2 = np.nanquantile(s2, 1.0 - percent_up)
        for key, val in data.data.items():
            if isinstance(key, (str)):
                break
            v = np.copy(val["s1"])
            v[v > vmax1] = np.nan
            val["s1"] = np.copy(v)
            if grad_data:
                v = np.copy(val["s2"])
                v[v > vmax2] = np.nan
                val["s2"] = np.copy(v)
        if grad_data:
            print(
                f"Clip above {np.round(vmax1,1)} for sensor 1 and "
                + f"{np.round(vmax2,1)} for sensor 2")
        else:
            print(f"Clip above {np.round(vmax1,1)}")
    data.sensor1, data.sensor2, _, _, _ = data_flatten(data.data)
    # self.main.data = data


def julian2date(j_day, year):
    """
    Function translates Julian day number to standard date.
    1st of January is Julian day number 1.

    Parameters
    ----------
    j_day : int
        Number of Julian day
    year : int
        Year in which to do the calculation (important to know whether
        it is a leap year). May be 2 or 4 ciphers

    Returns
    -------
    day: int
        Day of month
    month: int
        Month in year
    """
    day_month = np.array([0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304,
                          334])
    if year % 4 == 0:
        day_month[2:] += 1
    month = np.where(day_month >= j_day)[0][0]
    day = j_day - day_month[month - 1]
    return day, month


def date2julian(day, month, year):
    """
    Function translates month and day of month to Julian day of year.
    1st of January is Julian day number 1.

    Parameters
    ----------
    day : int
        Day of month (natural counting, starts at 1)
    month : int
        Month of year (natural counting, starts at 1 for January)
    year : int
        Year in which to do the calculation (important to know whether
        it is a leap year). May be 2 or 4 ciphers

    Returns
    -------
    j_day: int
        Julian day of year
    """
    day_month = np.array([0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304,
                          334])
    if year % 4 == 0:
        day_month[2:] += 1
    return day_month[month-1] + day


def time2data(time, year):
    """
    Convert seconds into julian day of year, hour, minute and second

    Parameters
    ----------
    time : may be a single float or a 1D numpy array
        Time to be converted [s], based on julian day of year, i.e.
        time = julian_day*86000+hour*3600+minute*60+second.

    Returns
    -------
    All returned parameters have the same shape as time.

    month :  int
        month of year
    day : int
        day of month.
    h : int
        Hour.
    m : int
        minute.
    s : float
        second.

    """
    try:
        _ = len(time)
        d = np.array(time / 86400.0, dtype=int)
        t = time - d * 86400.0
        h = np.array(t / 3600.0, dtype=int)
        t -= h * 3600.0
        m = np.array(t / 60.0, dtype=int)
        s = t - m * 60.0
        day = np.zeros_like(d)
        month = np.zeros_like(d)
        for i, dd in enumerate(d):
            day[i], month[i] = julian2date(dd, year)
    except TypeError:
        day = int(time / 86400.0)
        t = time - day * 86400.0
        h = int(t / 3600.0)
        t -= h * 3600.0
        m = int(t / 60.0)
        s = t - m * 60.0
        day, month = julian2date(day, year)
    return month, day, h, m, s


def diurnal_variation(times, data, degree=3):
    """
    Calculate a fit of degree "degree" to the data of lines "lines" which
    will be used for correction of diurnal variations if not base station
    data exist

    Parameters
    ----------
    times : 1D numpy float array
        Times for the data to be fitted in seconds
    data : 1D numpy float array
        Data to be fitted (usually a series of line medians)
    degree : int, optional
        Degree of polynom to be fitted to data. The default is 3.

    Returns
    -------
    Polynome coefficients (1D numpy array of size degree+1)
        The polynome is calculated as
        P[degree]+P[degree-1]*x+P[degree-2]*x**2...+P[0]*x**degree
        If multiple blocks are fitted together, P contains a polynome for
        each block, i.e. len(P) = (degree+1)*number_of_blocks
    tmn : float
        For the stability of polynome fit, times (given in seconds) are
        reduced such that the minimum time is zero. tmn is this minumum
        time. To apply the coefficients to data, their time must be
        transformed to time-tmn before applying the polynome coefficients.

    """
    # with open("test_data","w") as fo:
    #     for i,t in enumerate(times):
    #         fo.write(f"{t} {data[i]}\n")
    tmn = times.min()
    return np.polyfit(times-tmn, data, deg=degree), tmn


def diurnal_correction(data_c, base, base_flag=True, degree=5):
    """
    Apply diurnal corrections
    The diurnal variations may come from base station data (function
    read_base) or, if no base station data exist, they are calculated in
    function diurnal_variation by fitting a polynomial of degree "degree"
    to the measured data. Data of different days are then fitted
    independently.

    Base station data (measured or calculated) are interpolated onto
    measurement times and simply subtracted from data. The process is done
    in situ, i.e. the values of arrays self.sensor1 and self.sensor2 are
    modified. If you need to keep the original data, you must copy them to
    other arrays before applying diurnal_correction

    Parameters
    ----------
    data : Member of class data
        Contains data to be corrected
    base : Member of class Geometrics
        Constains base station data if there are or stores extimated ones
    base_flag : bool, optional; Default: True
        If True, base station data were read from file. Else, estimate
        diurnal variations from line medians
    degree : int, optional, Default: 3
        Degree of polynom to be fitted to data.
        This parameter is only used if no base station data exist

    """
    # Test whether base station data have already been read
    data = deepcopy(data_c.data)
# If no base station data exist, use median values of sensor 1 of every
#    line asindicator for diurnal variations.
# Add median value at the time of the beginning and the end of every line
    if not base_flag:
        month_base = []
        day_base = []
        year_base = []
        hour_base = []
        minute_base = []
        second_base = []
        time_base = []
        jday_base = []
        base_d = np.array([])
        year = data["year"]
        t = np.array([])
        v = np.array([])
        for key, val in data.items():
            if isinstance(key, (str)):
                break
            t = np.concatenate((t, val["time"]))
            v = np.concatenate((v, val["s1"]))
        index = np.where(np.isfinite(v))[0]
        t = t[index]
        v = v[index]
        days = np.array(t/86400., dtype=int)
        days_u = np.unique(days)
        params = []
        tm = []
        tdiurnal = np.array([])
        for d in days_u:
            index = np.where(days == d)
            fit_parameters, tmn = diurnal_variation(t[index], v[index],
                                                    degree=degree)
            params.append(fit_parameters)
            tm.append(tmn)
            t0 = t[index].min()
            t1 = t[index].max()
            tt = np.arange(t0-60., t1+100., 60.)-tmn
            npar = len(fit_parameters)
            vv = np.zeros(len(tt))
            for i in range(npar):
                vv += fit_parameters[i] * tt ** (degree - i)
            tdiurnal = np.concatenate((tdiurnal, tt+tmn))
            base_d = np.concatenate((base_d, vv))
        for tim in tdiurnal:
            mon, da, ho, mi, se = time2data(tim, year)
            year_base.append(year)
            month_base.append(mon)
            day_base.append(da)
            hour_base.append(ho)
            minute_base.append(mi)
            second_base.append(se)
            time_base.append(tim)
            jday_base.append(date2julian(int(da), int(mon), int(year)))
        month_base = np.array(month_base, dtype=int)
        day_base = np.array(day_base, dtype=int)
        jday_base = np.array(jday_base, dtype=int)
        year_base = np.array(year_base, dtype=int)
        hour_base = np.array(hour_base, dtype=int)
        minute_base = np.array(minute_base, dtype=int)
        second_base = np.array(second_base)
        time_base = np.array(time_base)
        with open("temp.stn", "w", encoding="utf-8") as fo:
            for i, b in enumerate(base_d):
                if np.isnan(b):
                    continue
                fo.write(
                    f"*  0 {jday_base[i]:3d} {hour_base[i]:02d}"
                    + f"{minute_base[i]:02d}"
                    + f"{int(second_base[i]):02d}"
                    + f"{i:5d}{int(b*10):7d}\n")
        base.read_base("temp.stn", year)
        os.remove("temp.stn")
    else:
        time1 = []
        time2 = []
        for key, val in data.items():
            if isinstance(key, (str)):
                break
            time1.append(min(val["time"][0], val["time"][-1]))
            time2.append(max(val["time"][0], val["time"][-1]))
        time1 = np.array(time1)
        time2 = np.array(time2)
# Check whether base station data cover all measurements.
# If not, set base station data to zero and avoid in this way base
# corrections
        if base.time_base.min() > time1.min() or\
                base.time_base.max() < time2.max():
            print("\nWARNING in function  diurnal_correction:")
            print("   Base station data do not cover all measurements.")
            print("           No base station corrections effectuated.\n")
            _ = QtWidgets.QMessageBox.warning(
                None, "Warning",
                "Function utilities.diurnal_correction:\n\n"
                + "Base station data do not cover all measurements\n"
                + f"Base station between day {base.time_base.min()/86400.} "
                + f"and {base.time_base.max()/86400.}\n"
                + "Instrument data between day "
                + f"{time1.min()/86400.} and {time2.max()/86400.}\n\n"
                + "No base station corrections effectuated.",
                QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
            return False
# interpolate base station values at measurement times
    tb, ind = np.unique(base.time_base, return_index=True)
    b = base.base[ind]
    f = interpolate.interp1d(tb, b, kind=1)
    for key, val in data.items():
        if isinstance(key, (str)):
            break
        diurnal = f(val["time"])
# Subtract base station data from field data
        val["s1"] -= diurnal
        val["median1"] = np.nanmedian(val["s1"])
        if data["grad_data"]:
            val["s2"] -= diurnal
            val["median2"] = np.nanmedian(val["s2"])
    data_c.data = deepcopy(data)
    return True


def interpol_line(data, nsensor, i_line=0, dx=0.2, xmin=0.0, xmax=0.0, k=3):
    """
    interpolate data of one line onto a regular grid

    Parameters
    ----------
    data : Member of class data
        data to be interpolated
    i_line : int, optional. Default is 0.
        Number of line to be interpolated (counting starts at 0).
    dx : float, optional
        Sampling step in meters for interpolated data. The default is 0.2.
    xmin : float, optional. Default is 0.
        Position of first sample along self.direction in meters.
    xmax : float, optional. Default is 0.
        Position of last sample along self.direction in meters.
    k : int, optional
        Degree of spline used for interpolation. The default is 3.
        See scipy.interpolate.interp1d for more information.
        Only splines are used. Correspondance between k and "kind" of
        scipy.interpolate.interp1d:

        - k=0: kind="zero"
        - k=1: kind="slinear"
        - k=2: kind="quadratic"
        - k=3: kind="cubic"

    If xmin == xmax, the starting and end points are calculated
    automatically. For this,the starting point is placed at the nearest
    multiple of dx for the coordinate of self.direction (see function
    read_stn)

    Returns
    -------
    sensor_inter: numpy float array
        Interpolated data
    x_inter: numpy float array
        Interpolated X-coordinates
    y_inter: numpy float array
        Interpolated Y-coordinates
    dmin: float
        Position of first interpolated point within line [m]
    dmax: float
        Position of last interpolated point within line [m]

    """
    kind = ["zero", "slinear", "quadratic", "cubic"]
    if k > 3 or k < 0:
        print("\nWARNING Function interpol_line:\n"
              f"      Given k ({k}) is not allowed. k is set to 3 (cubic)")
        k = 3
    # Extract data
    data = data.data
    if nsensor == 1:
        s1 = data[i_line]["s1"]
    else:
        s1 = data[i_line]["s2"]
    index = np.isfinite(s1)
    s1 = s1[index]
    xdat = data[i_line]["x"]
    xdat = xdat[index]
    ydat = data[i_line]["y"]
    ydat = ydat[index]
# Define coordinates in principal direction along which to interpolate data
    if data["direction"] in ("N", "S", 0., 180.):
        ddat = np.copy(xdat)
    else:
        ddat = np.copy(ydat)
# Define starting and end points
    if xmin == xmax:
        dmin = np.ceil(np.round(ddat.min(), 2) / dx) * dx
        dmax = np.floor(np.round(ddat.max(), 2) / dx) * dx
    else:
        dmin = xmin
        dmax = xmax
# Calculate number of interpolated data and their positions along the principal
#   direction
    nx = np.int((dmax - dmin) / dx + 1)
    d_inter = dx * np.arange(nx) + dmin
# Do interpolation for first sensor
    f = interpolate.interp1d(ddat, s1, kind=kind[k],
                             fill_value="extrapolate")
    sensor_inter = f(d_inter)
# Do interpolation for X-coordinates
    f = interpolate.interp1d(ddat, xdat, kind=kind[k],
                             fill_value="extrapolate")
    x_inter = f(d_inter)
    # Do interpolation for Y-coordinates
    f = interpolate.interp1d(ddat, ydat, kind=kind[k],
                             fill_value="extrapolate")
    y_inter = f(d_inter)
    return sensor_inter, x_inter, y_inter, dmin, dmax


def interpol_2D(data_c, dx=0.2, dy=0.2):
    """
    Routine interpolates data on all lines onto a regular grid. No
    extrapolation is done, i.e. if at the beginning or the end of a line
    data are missing (the line starts later than others or stops earlier),
    the interpolated array will contain nans
    The interpolation method used is
    scipy.interpolate.CloughTocher2DInterpolator

    Parameters
    ----------
    data_c : member of class data
        data to be interpolated
    dx : float, optional
        Sampling step in meters in x-direction. The default is 0.2.
    dx : float, optional
        Sampling step in meters in y-direction. The default is 0.2.

    Returns
    -------
    sensor1_inter : 2D numpy float array
        Contains gridded data of sensor 1
    sensor2_inter : 2D numpy float array
        Contains gridded data of sensor 2
    grad_inter :2D numpy float array
        contains the vertical gradient

    The shape of the arrays depends on the principal direction of the
    lines:
    - If self.direction == 1, shape is (number_of_data_points_per_line,
      number_of_lines)
    - else: (number_of_lines, number_of_data_points_per_line)

    x_inter : 1D numpy float array
        x_coordinates of the columns of s1_inter and s2_inter
    y_inter : 1D numpy float array
        y_coordinates of the rows of s1_inter and s2_inter

    """
    data = data_c.data
    keys = []
    for key in data.keys():
        if isinstance(key, (str)):
            break
        keys.append(int(key))
    keys = np.array(keys)
    key1 = keys[0]
    direction = data[key1]["direction"]
    if direction in ("N", "S", 0., 180.):
        direction = "N"
    else:
        direction = "E"
    if data_c.grad_data:
        nsensor = 2
    else:
        nsensor = 1
# search all line positions. If different blocks were joint, it is possible
# that pieces of one the same line were measured in different blocks and these
# pieces should be joint before interpolation
    pos_l = []
    for k in keys:
        if direction == "N":
            pos_l.append(np.nanmedian(data[k]["x"]))
        else:
            pos_l.append(np.nanmedian(data[k]["y"]))
    pos_l = np.array(pos_l)
    line_positions = np.unique(pos_l)
    nl = len(line_positions)
# Define grid area covering the whole measured area
    xmin = data[key1]["x"].min()
    xmax = data[key1]["x"].max()
    ymin = data[key1]["y"].min()
    ymax = data[key1]["y"].max()
    for key, val in data.items():
        if isinstance(key, (str)):
            break
        xmin = min(val["x"].min(), xmin)
        xmax = max(val["x"].max(), xmax)
        ymin = min(val["y"].min(), ymin)
        ymax = max(val["y"].max(), ymax)
    xmin = np.ceil(xmin/dx) * dx
    xmax = np.floor(xmax/dx) * dx
    ymin = np.ceil(ymin/dy) * dy
    ymax = np.floor(ymax/dy) * dy
    x_inter = xmin + np.arange(int((xmax - xmin) / dx) + 1) * dx
    y_inter = ymin + np.arange(int((ymax - ymin) / dy) + 1) * dy
    xi, yi = np.meshgrid(x_inter, y_inter)
    ny, nx = xi.shape
# set arrays for interpolation and predefine certain values to nan
    if direction == "N":
        posx = np.zeros((ny, nl))
        p_inter = y_inter
        p2_inter = x_inter
        nin = ny
        ncross = nx
        dp = dy
        dp2 = dx
    else:
        posx = np.zeros((nx, nl))
        p_inter = x_inter
        p2_inter = y_inter
        nin = nx
        ncross = ny
        dp = dx
        dp2 = dy
    zint = np.zeros_like(posx)
    s1int = np.zeros_like(posx)
    s2int = np.zeros_like(posx)
    tim = np.zeros_like(posx)
    posx[:, :] = np.nan
    zint[:, :] = np.nan
    s1int[:, :] = np.nan
    s2int[:, :] = np.nan
    for iline, lpos in enumerate(line_positions):
        x = pos = z = t = s1 = s2 = np.array([])
        for key, val in data.items():
            if isinstance(key, (str)):
                break
            if direction == "N":
                if not np.isclose(np.nanmedian(val["x"]), lpos):
                    continue
                x = np.concatenate((x, val["x"]))
                pos = np.concatenate((pos, val["y"]))
            else:
                if not np.isclose(np.nanmedian(val["y"]), lpos):
                    continue
                x = np.concatenate((x, val["y"]))
                pos = np.concatenate((pos, val["x"]))
            z = np.concatenate((z, val["z"]))
            t = np.concatenate((t, val["time"]))
            s1 = np.concatenate((s1, val["s1"]))
            if nsensor == 2:
                s2 = np.concatenate((s2, val["s2"]))
        pos, pos_ind = np.unique(pos, return_index=True)
        x = x[pos_ind]
        z = z[pos_ind]
        s1 = s1[pos_ind]
        t = t[pos_ind]
        index = np.where(np.isfinite(s1))[0]
        pi = pos[index]
        pm = pos.min()
        n1 = np.where(p_inter >= pm)[0][0]
        pm = pos.max()
        n2 = np.where(p_inter <= pm)[0][-1] + 1
        f = interpolate.interp1d(pi, s1[index], kind="quadratic")
        s1int[n1:n2, iline] = f(p_inter[n1:n2])
        f = interpolate.interp1d(pi, x[index], kind="linear")
        posx[n1:n2, iline] = f(p_inter[n1:n2])
        f = interpolate.interp1d(pi, z[index], kind="quadratic")
        zint[n1:n2, iline] = f(p_inter[n1:n2])
        f = interpolate.interp1d(pi, t[index], kind="linear")
        tim[n1:n2, iline] = f(p_inter[n1:n2])

# Check whether there are big holes in data along the line (>10*step size)
        hole = np.where(abs(pi[1:]-pi[:-1]) > 10.*dp)[0]
        if len(hole > 0):
            for h in hole:
                n1 = int((pi[h]-p_inter[0])/dp)
                n2 = int((pi[h+1]-p_inter[0])/dp)+1
                s1int[n1:n2, iline] = np.nan
                posx[n1:n2, iline] = np.nan
                zint[n1:n2, iline] = np.nan
        if nsensor == 2:
            s2 = s2[pos_ind]
            index = np.where(np.isfinite(s2))[0]
            pi = pos[index]
            pm = pi.min()
            n1 = np.where(p_inter >= pm)[0][0]
            pm = pi.max()
            n2 = np.where(p_inter <= pm)[0][-1] + 1
            f = interpolate.interp1d(pi, s2[index], kind="quadratic")
            s2int[n1:n2, iline] = f(p_inter[n1:n2])
            hole = np.where(abs(pi[1:]-pi[:-1]) > 10.*dp)[0]
            if len(hole > 0):
                for h in hole:
                    n1 = int((pi[h]-p_inter[0])/dp)
                    n2 = int((pi[h+1]-p_inter[0])/dp)+1
                    s2int[n1:n2, iline] = np.nan
    z_inter = np.zeros((ncross, nin))
    z_inter[:, :] = np.nan
    s1_inter = np.zeros_like(z_inter)
    s1_inter[:, :] = np.nan
    s2_inter = np.zeros_like(z_inter)
    s2_inter[:, :] = np.nan
    t_inter = np.zeros_like(z_inter)
    for i in range(len(p_inter)):
        pos, pos_ind = np.unique(posx[i, :], return_index=True)
        s1 = s1int[i, :][pos_ind]
        t = tim[i, :][pos_ind]
        index = np.where(np.isfinite(s1))[0]
        if len(index) < 1:
            continue
        i1 = index[0]
        i2 = index[-1] + 1
        pi = pos[i1:i2]
        pm = pi.min()
        n1 = np.where(p2_inter >= pm)[0][0]
        pm = pi.max()
        n2 = np.where(p2_inter <= pm)[0][-1] + 1
        pi = pi[index]
        s1 = s1[index]
        f = interpolate.interp1d(pi, s1[i1:i2], kind="quadratic")
        s1_inter[n1:n2, i] = f(p2_inter[n1:n2])
        z = zint[i, :]
        z = z[pos_ind]
        f = interpolate.interp1d(pi, z[i1:i2], kind="quadratic")
        z_inter[n1:n2, i] = f(p2_inter[n1:n2])
        f = interpolate.interp1d(pi, t[i1:i2], kind="quadratic")
        t_inter[n1:n2, i] = f(p2_inter[n1:n2])
# Check whether there are big holes in data along the line (>3*line distance)
        if len(index) > 1:
            hole = np.where(pos_ind[1:]-pos_ind[:-1] > 3)[0]
            if len(hole > 0):
                for ih, h in enumerate(hole):
                    n1 = int((pos[pos_ind[hole[0]]]-p2_inter[0])/dp2)+1
                    n2 = int((pos[pos_ind[hole[0]]+1]-p2_inter[0])/dp2)
                    s1_inter[n1:n2, i] = np.nan
                    z_inter[n1:n2, i] = np.nan
        if nsensor == 2:
            s2 = s2int[i, :][pos_ind]
            index = np.where(np.isfinite(s2))[0]
            i1 = index[0]
            i2 = index[-1] + 1
            pi = pos[i1:i2]
            pm = pi.min()
            n1 = np.where(p2_inter >= pm)[0][0]
            pm = pos[i1:i2].max()
            n2 = np.where(p2_inter <= pm)[0][-1] + 1
            pi = pi[index]
            s2 = s2[index]
            f = interpolate.interp1d(pi, s2[i1:i2], kind="quadratic")
            s2_inter[n1:n2, i] = f(p2_inter[n1:n2])
            if len(index) > 1:
                hole = np.where(pos_ind[1:]-pos_ind[:-1] > 3)[0]
                if len(hole > 0):
                    for ih, h in enumerate(hole):
                        n1 = int((pos[pos_ind[hole[0]]]-p2_inter[0])/dp2)
                        n2 = int((pos[pos_ind[hole[0]]+1]-p2_inter[0])/dp2)
                        s2_inter[n1+1:n2, i] = np.nan
    if data[key1]["direction"] in ("N", "S", 0., 180.):
        s1_inter = s1_inter.T
        z_inter = z_inter.T
        t_inter = t_inter.T
        if nsensor == 2:
            s2_inter = s2_inter.T
    if nsensor == 2:
        grad_inter = (s2_inter-s1_inter)/data["d_sensor"]
        return s1_inter, s2_inter, grad_inter, x_inter, y_inter, z_inter, \
            t_inter
    dum = 0
    return s1_inter, dum, dum, x_inter, y_inter, z_inter, t_inter


def extrapolate(d, x, y):
    """
    Routine fills nans on an interpolated grid.
    For this, it searches first for every line and column the first and
    last existing (non-nan) points. Then, for every non-defined point, it
    searches the "n_nearest" nearest points (see first command line) and
    associates a weighted average value. The weight is calculated as
    1/distance**2

    Parameters
    ----------
    d : 2D numpy array, shape: (ny,nx)
           Contains data on regular grid. NaN for inexistant data.
    x : 1D numpy array
        Coordiantes of the columns of data
    y : 1D numpy array
        Coordiantes of the rows of data

    Returns
    -------
    data: 2D numpy array with the same shape as input data.
        Contains full regular grid of data

    """
    n_nearest = 8
    data = np.copy(d)
    XX, YY = np.meshgrid(x, y)
    # Search all points at the beginning and the end of all rows
    zlim = []
    xlim = []
    ylim = []
    for i in range(len(x)):
        j1 = len(y)
        j2 = 0
        for j in range(len(y)):
            if np.isfinite(data[j, i]):
                j1 = j
                break
        for j in range(len(y) - 1, -1, -1):
            if np.isfinite(data[j, i]):
                j2 = j
                break
        if j1 <= j2:
            zlim.append(data[j1, i])
            xlim.append(XX[j1, i])
            ylim.append(YY[j1, i])
            zlim.append(data[j2, i])
            xlim.append(XX[j2, i])
            ylim.append(YY[j2, i])
    # Search all points at the beginning and the end of all columns
    for j in range(len(y)):
        i1 = len(x)
        i2 = 0
        for i in range(len(x)):
            if np.isfinite(data[j, i]):
                i1 = i
                break
        for i in range(len(x) - 1, -1, -1):
            if np.isfinite(data[j, i]):
                i2 = i
                break
        if i1 <= i2:
            zlim.append(data[j, i1])
            xlim.append(XX[j, i1])
            ylim.append(YY[j, i1])
            zlim.append(data[j, i2])
            xlim.append(XX[j, i2])
            ylim.append(YY[j, i2])
    zlim = np.array(zlim)
    xlim = np.array(xlim)
    ylim = np.array(ylim)
    # Do extrapolation
    for i, xx in enumerate(x):
        for j, yy in enumerate(y):
            if np.isnan(data[j, i]):
                dist = 1 / ((xx - xlim) ** 2 + (yy - ylim) ** 2)
                ind = np.argsort(dist)
                data[j, i] = np.dot(
                    dist[ind[-n_nearest:]], zlim[ind[-n_nearest:]])\
                    / np.sum(dist[ind[-n_nearest:]])
    return data


def justify_lines_median(data_c, just=0, inplace=True):
    """
    Often the measurment direction has an influence on magnetic data due to
    uncorrected effects of acquisition instrumentation.
    The function calculates the median values of every line and adjusts the
    one of every second line to the average median of the neighbouring
    lines

    Parameters
    ----------
    just : int, optional. Default is 0.
        If 0: Leave medians of even line (python counting, i.e. first line
        is even) untouched, justify odd lines to medians of even lines.
        If 1: Do the reverse
    inplace : bool, optional. Default is True
        if True, justified data are back-copied to self.sensorN_inter and
        True is returned. If not, new arrays are returned.

    Returns
    -------
    s1_justified : 1D numpy array with justified data of first sensor
    s2_justified : 1D numpy array with justified data of second sensor

    """
    data = deepcopy(data_c.data)
    keys = []
    for key in data.keys():
        if isinstance(key, str):
            break
        keys.append(key)
    keys = np.array(keys)
    max_key = keys.max()
    if just == 0:
        d_change = data[1]["direction"]
        d_keep = data[0]["direction"]
    else:
        d_change = data[0]["direction"]
        d_keep = data[1]["direction"]
        # for day in self.lines_per_day.keys():
    for key, val in data.items():
        k = keys[1]
        k1 = keys[1]
        k2 = keys[1]
        if isinstance(key, str):
            break
        if val["direction"] == d_change:
            if key == 0:
                for k in keys[1:]:
                    if data[k]["direction"] == d_keep:
                        break
                dm1 = data[key]["median1"] - data[k]["median1"]
                if data["grad_data"]:
                    dm2 = data[key]["median2"] - data[k]["median2"]
            elif key == max_key:
                for k in keys[-1::-1] == d_keep:
                    break
                dm1 = data[key]["median1"] - data[k]["median1"]
                if data["grad_data"]:
                    dm2 = data[key]["median2"] - data[k]["median2"]
            else:
                for k1 in keys[key+1:]:
                    if data[k1]["direction"] == d_keep:
                        break
                for k2 in keys[key-1::-1]:
                    if data[k2]["direction"] == d_keep:
                        break
                dm1 = (data[key]["median1"]
                       - (data[k1]["median1"] + data[k2]["median1"]) / 2.0)
                if data["grad_data"]:
                    dm2 = (data[key]["median2"]
                           - (data[k1]["median2"] + data[k2]["median2"]) / 2.0)
            val["s1"] -= dm1
            val["median1"] -= dm1
            if data["grad_data"]:
                val["s2"] -= dm2
                val["median2"] -= dm2
    if inplace:
        data_c.data = deepcopy(data)
        return True
    return data


def gauss_transform(data_fix, data_move):
    """
    Transforms data sets to gaussian distribution does a projection
    of the second data set onto the distribution of the first and returns
    the back-transformed modified second data set

    Parameters
    ----------
    data_fix : numpy 1D array
        Reference data set.
    data_move : numpy 1D array
        Data set to be projected onto the gaussian distribution of
        data_fix.

    Returns
    -------
    numpy 1D array
        Modified data_move array.

    """
# For the number of quantiles take the number of data of the smaller data set
    n = min(len(data_fix), len(data_move))
# It seems that the number of quantiles in
# sklearn.preprocessing.QuantileTransformer is limited to 10000.
    n = min(n, 10000)
# Do the Gauss-transform of the reference data set
    GT_fix = QuantileTransformer(n_quantiles=n,
                                 output_distribution="normal")
    _ = GT_fix.fit_transform(data_fix)[:, 0]
# Do the Gauss-transform of the data set to be modified
    GT_move = QuantileTransformer(n_quantiles=n,
                                  output_distribution="normal")
    v_move = GT_move.fit_transform(data_move)
# Project data_move onto the Gauss distribution of data_fix and return the
#     back-transformed data.
    return GT_fix.inverse_transform(v_move)[:, 0]


def justify_lines_gaussian(data, just=0, local=1, inplace=True):
    """
    Often the measurment direction has an influence on magnetic data due to
    uncorrected effects of acquisition instrumentation.
    The function calculates the median values of every line and adjusts the
    one of every second line to the average median of the neighbouring
    lines

    Parameters
    ----------
    just : int, optional. Default is 0.
        If 0: Leve medians of even line (python counting, i.e. first line
        is even) untouched, justify odd lines to medians of even lines
        If 1: Do the reverse
    local : int, optional. Default is 1
        If 0: apply gaussian transform to the whole data set
        If 1: apply gaussian transform only to neighboring lines
    inplace : bool, optional. default is True
        If True, justified data are back-copied to self.sensorN_inter and
        True is returned. If not, new arrays are returned

    Returns
    -------
    s1_justified : 2D numpy float array
        Justified data of first sensor
    s2_justified : 2D numpy float array
        Justified data of second sensor

    """
    s1_justified = np.copy(data.sensor1_inter)
    s2_justified = np.copy(data.sensor2_inter)

    if local:
        if just == 0:
            if data.direction == 0:
                nlines = data.sensor1_inter.shape[1]
                for i in range(1, nlines, 2):
                    data_fix = data.sensor1_inter[:, i-1:i+2:2]\
                        .reshape(-1, 1)
                    data_move = data.sensor1_inter[:, i].reshape(-1, 1)
                    s1_justified[:, i] = gauss_transform(data_fix, data_move)
                    data_fix = data.sensor2_inter[:, i-1:i+2:2]\
                        .reshape(-1, 1)
                    data_move = data.sensor2_inter[:, i].reshape(-1, 1)
                    s2_justified[:, i] = gauss_transform(data_fix, data_move)
            else:
                nlines = data.sensor1_inter.shape[0]
                for i in range(1, nlines, 2):
                    data_fix = data.sensor1_inter[i-1:i+2:2, :].reshape(-1, 1)
                    data_move = data.sensor1_inter[i, :].reshape(-1, 1)
                    s1_justified[i, :] = gauss_transform(data_fix,
                                                         data_move)
                    data_fix = data.sensor2_inter[i-1:i+2:2, :].reshape(-1, 1)
                    data_move = data.sensor2_inter[i, :].reshape(-1, 1)
                    s2_justified[i, :] = gauss_transform(data_fix, data_move)
        else:
            if data.direction == 0:
                nlines = data.sensor1_inter.shape[1]
                data_fix = data.sensor1_inter[:, 1].reshape(-1, 1)
                data_move = data.sensor1_inter[:, 0].reshape(-1, 1)
                s1_justified[:, 0] = gauss_transform(data_fix, data_move)
                data_fix = data.sensor2_inter[:, 1].reshape(-1, 1)
                data_move = data.sensor2_inter[:, 0].reshape(-1, 1)
                s2_justified[:, 0] = gauss_transform(data_fix, data_move)
                for i in range(2, nlines, 2):
                    data_fix = data.sensor1_inter[:, i-1:i+2:2].reshape(-1, 1)
                    data_move = data.sensor1_inter[:, i].reshape(-1, 1)
                    s1_justified[:, i] = gauss_transform(data_fix, data_move)
                    data_fix = data.sensor2_inter[:, i-1:i+2:2].reshape(-1, 1)
                    data_move = data.sensor2_inter[:, i].reshape(-1, 1)
                    s2_justified[:, i] = gauss_transform(data_fix, data_move)
            else:
                nlines = data.sensor1_inter.shape[0]
                data_fix = data.sensor1_inter[1, :].reshape(-1, 1)
                data_move = data.sensor1_inter[0, :].reshape(-1, 1)
                s1_justified[0, :] = gauss_transform(data_fix, data_move)
                data_fix = data.sensor2_inter[1, :].reshape(-1, 1)
                data_move = data.sensor2_inter[0, :].reshape(-1, 1)
                s2_justified[0, :] = gauss_transform(data_fix, data_move)
                for i in range(2, nlines, 2):
                    data_fix = data.sensor1_inter[i-1:i+2:2, :].reshape(-1, 1)
                    data_move = data.sensor1_inter[i, :].reshape(-1, 1)
                    s1_justified[i, :] = gauss_transform(data_fix, data_move)
                    data_fix = data.sensor2_inter[i-1:i+2:2, :].reshape(-1, 1)
                    data_move = data.sensor2_inter[i, :].reshape(-1, 1)
                    s2_justified[i, :] = gauss_transform(data_fix, data_move)
    else:
        if just == 0:
            if data.direction == 0:
                s = gauss_transform(data.sensor1_inter[:, 0::2].reshape(-1, 1),
                                    data.sensor1_inter[:, 1::2].reshape(-1, 1))
                s1_justified[:, 1::2] = s.reshape(
                    data.sensor1_inter[:, 1::2].shape)
                s = gauss_transform(data.sensor2_inter[:, 0::2].reshape(-1, 1),
                                    data.sensor2_inter[:, 1::2].reshape(-1, 1))
                s2_justified[:, 1::2] = s.reshape(
                    data.sensor2_inter[:, 1::2].shape)
            else:
                s = gauss_transform(data.sensor1_inter[0::2, :].reshape(-1, 1),
                                    data.sensor1_inter[1::2, :].reshape(-1, 1))
                s1_justified[1::2, :] = s.reshape(
                    data.sensor1_inter[1::2, :].shape)
                s = gauss_transform(data.sensor2_inter[0::2, :].reshape(-1, 1),
                                    data.sensor2_inter[1::2, :].reshape(-1, 1))
                s2_justified[1::2, :] = s.reshape(
                    data.sensor1_inter[1::2, :].shape)
        else:
            if data.direction == 0:
                s = gauss_transform(data.sensor1_inter[:, 1::2].reshape(-1, 1),
                                    data.sensor1_inter[:, 0::2].reshape(-1, 1))
                s1_justified[:, 0::2] = s.reshape(
                    data.sensor1_inter[:, 1::2].shape)
                s = gauss_transform(data.sensor2_inter[:, 1::2].reshape(-1, 1),
                                    data.sensor2_inter[:, 0::2].reshape(-1, 1))
                s2_justified[:, 0::2] = s.reshape(
                    data.sensor2_inter[:, 1::2].shape)
            else:
                s = gauss_transform(data.sensor1_inter[1::2, :].reshape(-1, 1),
                                    data.sensor1_inter[0::2, :].reshape(-1, 1))
                s1_justified[0::2, :] = s.reshape(
                    data.sensor1_inter[1::2, :].shape)
                s = gauss_transform(data.sensor2_inter[1::2, :].reshape(-1, 1),
                                    data.sensor2_inter[0::2, :].reshape(-1, 1))
                s2_justified[0::2, :] = s.reshape(
                    data.sensor1_inter[1::2, :].shape)

    if inplace:
        data.sensor1_inter = np.copy(s1_justified)
        data.sensor2_inter = np.copy(s2_justified)
        return True
    return s1_justified, s2_justified


def next_minute(hour, minute, second):
    """
    Get next full minute after actual time

    Parameters
    ----------
    hour : int
        Actual hour
    minute : int
        Actual minute
    second : float
        Actual second

    Returns
    -------
    hour : int
        hour of next minute
    minute : int
        next minute
    second : float
        0.
    """
    second = 0.
    minute += 1
    if minute == 60:
        hour += 1
        minute = 0
    return hour, minute, second


def add_time(hour, minute, second, dt):
    """
    Add a time step to actual time

    Parameters
    ----------
    hour : int
        Actual hour
    minute : int
        Actual minute
    second : float
        Actual second
    dt : Time to be added in seconds

    Returns
    -------
    hour : int
        hour of new time
    minute : int
        minute of new time
    second : float
        second of new time
    """
    second += dt
    if second == 60.:
        minute += 1
        second = 0.
        if minute == 60:
            hour += 1
            minute = 0
    return hour, minute, second


def fit2lines(x, y, n0, n1, n2, n0_flag):
    """
    Fit two regression lines to a data set.
    Find the point where a break in slope gives the best fit between the
    n1th point and the n2th point

    Parameters
    ----------
    x : numpy 1D array, float
        x-coordinates of the data series.
    y : numpy 1D array, float
        y-coordinates of the data series.
    n0 : int
        first point of the series to be considered.
    n1 : int
        First point in series considered for possible slope break.
    n2 : int
        Last point in series considered for possible slope break.
    n0_flag : bool
        if True : first line must pass through point n0

    Returns
    -------
    regression coefficients for first slope
    regression coefficients for second slope
    int: position of slope beak
    float: misfit

    """
    r1_best = None
    r2_best = None
    qual = 1.0e20
    n3 = len(x)
    isplit = n1
    slopes1 = []
    slopes2 = []
    inter1 = []
    inter2 = []
    fits = []
    isplits = []
# Fit two regression lines to data. For this, search breaking point between
#     third and 11th data point for which the fit is best
    for i in range(n1, n2):
        k1 = x[n0:i+1].reshape(-1, 1)
        k2 = x[i:].reshape(-1, 1)
        if n0_flag:
            reg1 = LR(fit_intercept=False).fit(k1-k1[0], y[n0:i+1]-y[n0])
            reg1.intercept_ = y[n0] - k1[0][0] * reg1.coef_[0]
        else:
            reg1 = LR(fit_intercept=True).fit(k1, y[n0:i+1])
        reg2 = LR(fit_intercept=True).fit(k2, y[i:])
        yy = np.zeros(n3)
        yy[n0:i+1] = k1.flatten()*reg1.coef_[0]+reg1.intercept_-y[n0:i+1]
        yy[i:] = k2.flatten()*reg2.coef_[0]+reg2.intercept_-y[i:]
        fit = np.sum(yy**2)
        # fit = reg1.score(k1, y[n0:i+1])/(i-n0)\
        #     + reg2.score(k2, y[i:])/(n3-i)
# If fit is better than earlier ones, calculate depths from both slopes
        slopes1.append(reg1.coef_[0])
        slopes2.append(reg2.coef_[0])
        inter1.append(reg1.intercept_)
        inter2.append(reg2.intercept_)
        fits.append(fit)
        isplits.append(i)
        if np.isfinite(fit) and fit < qual:
            qual = fit
            r1_best = reg1
            r2_best = reg2
            isplit = i
    return r1_best, r2_best, isplit, qual, slopes1, slopes2, inter1, inter2, \
        fits, isplits


def min_max(data, half_width=3):
    """
    Find all relative minima and maxima in a data vector.

    A maximum is found if a value at position i of the vector is larger
    than or equal to all other values in a range [i-half_width:i+half_with]
    and at the same time strictly larger than all values of one side.
    Sometimes, e.g. if seismograms are saturated, a slope exists on one
    side, but the values are constant on the other side. The one-sided test
    is to avoid that a flat curve with 2*half_width constant values is also
    considered as local maximum. Equivalent scheme for definition of a
    minimum. In addition, the function reduces possible local maxima and
    minima such that a maximum is always followed by a minimum and vice
    versa. If several local maxima follow each other (i.e. no wide enough
    local minimum exists between them), the program searches the strongest
    one of the subsequent maxima or, if several equal maximum values exist,
    it takes as position of the maximum the center point between those
    multiple maxima (again for saturated curves).

    Parameters
    ----------
    data : 1D numpy float array
        Data vrctor to be analysed
    half_width : int, optional (default: 3)
        Number of samples analyzed to all ides of every data sample.

    Returns
    -------
    max_pos: 1D numpy int array
        All position numbers where there is a relative maximum in vector
        "data".
    max_val: 1D numpy float array
        Values at these positions
    min_pos: 1D numpy int array
        All position numbers where there is a relative minimum in vector
        "data"
    min_val: 1D numpy float array
        Values at these positions
    """
    N = len(data)
    NN = np.arange(N, dtype="int")
# extreme_pos (extreme_neg) will contain the sum of all values <= (>=) the
#   central value
# half will contain the maximum of the number of
#   values < (>) the central value on the left and the right side
# A maximum (minimum) is found if extreme_xxx[i]==(2*half_width+1) and if
#   half_extreme_xxx[i]==half_width.
    extreme_pos = np.zeros(N, dtype=bool)
    extreme_neg = np.zeros(N, dtype=bool)
# Start loop over data points
# Sum of neigbouring points for which value[i] <= (>=) value[test_point]
    for k in range(N):
        dn0 = min(half_width, k)
        dn1 = min(half_width, N - k - 1) + 1
        width = dn0 + dn1
        ext_pos = sum(data[k] - data[k-dn0:k+dn1] >= 0) == width
# Sum of neighbouring values to the left (half1) and right (half2) < value
# [test_point]
        half1 = sum(data[k] - data[k-dn0:k] > 0)
        half2 = sum(data[k] - data[k+1:k+dn1] > 0)
        half = max(half1, half2) == half_width
        extreme_pos[k] = ext_pos and half

# Sum of neighbouring values to the left (half1) or right (half2) > value
# [test_point]
        ext_neg = sum(data[k] - data[k-dn0:k+dn1] <= 0) == width
        half1 = sum(data[k] - data[k-dn0:k] < 0)
        half2 = sum(data[k] - data[k+1:k+dn1] < 0)
        half = max(half1, half2) == half_width
        extreme_neg[k] = ext_neg and half
# Search all points that fulfill the criteria for local maximum and minimum
#        max_pos = NN[(extreme_pos==width) & (half_extreme_pos==half_width)]
    max_pos = NN[extreme_pos]
    max_val = data[max_pos]
#        min_pos = NN[(extreme_neg==width) & (half_extreme_neg==half_width)]
    min_pos = NN[extreme_neg]
    min_val = data[min_pos]
    del extreme_pos, extreme_neg
# mx_sig is a vector with length equal to number of found maxima with +1
# mn_sig is a vector with length equal to number of found maxima with -1
#   These vectors will be used to know which position is a maximum, which one
#   a minimum, once all extrema are concatenated in a single vector in order to
#   intercalate maxima and minima and to find places where multiple maxima or
#   minima follow each other
    mx_sig = np.ones(len(max_pos))
    mn_sig = -np.ones(len(min_pos))
# Concatenate positions, values and signs of maxima and minima into a single
#   vector for each of them
    signs = np.concatenate((mx_sig, mn_sig))
    positions = np.concatenate((max_pos, min_pos))
    values = np.concatenate((max_val, min_val))
# Order the concatenated vectors by positions
    iord = np.argsort(positions)
    pord = positions[iord]
    vord = values[iord]
    sord = signs[iord]
    ls = len(sord)
# Prepare lists that will contain positions, values and signs of alternating
#   extreme values (avoiding having several maxima (minima) following each
#   other without a minumum (maximum) between them).
    pos = []
    val = []
    sig = []
    i = 1
    # Start loop over concatenated extreme positions
# If sign of position [i] is different from position [i-1] accept position
#   [i-1] into a new list
    while i < ls:
        if sord[i] != sord[i - 1]:
            pos.append(pord[i - 1])
            val.append(vord[i - 1])
            sig.append(sord[i - 1])
        if i == ls - 1:
            if sord[i] != sord[i - 1]:
                pos.append(pord[i])
                val.append(vord[i])
                sig.append(sord[i])
                i += 1
            break
        #            continue
# if sign of position i is the same as the one of position i-1 search for next
#   position where sign changes
        i1 = i + 1
        for i1 in range(i + 1, ls):
            if sord[i] != sord[i1]:
                break
        if i1 < i:
            break
# Search maximum values of the positions having the same sign
#   the chosen position is the average position of all equal maximum (minimum)
#   values. If one of the relative maxima (minima) has the strongest value, its
#   position and value will be copied into the new list.
        if sord[i] > 0:
            mx = np.where(vord[i:i1] == max(vord[i:i1]))
            mpos = int(np.mean(pord[i:i1][mx]))
            pos.append(mpos)
            val.append(max(vord[i:i1]))
            sig.append(sord[i])
        else:
            mx = np.where(vord[i:i1] == min(vord[i:i1]))
            mpos = int(np.mean(pord[i:i1][mx]))
            pos.append(mpos)
            val.append(min(vord[i:i1]))
            sig.append(sord[i])
        i = i1 + 1
    del max_pos, max_val, min_pos, min_val, iord, pord, vord, sord
# Transform lists to numpy arrays
    pos = np.array(pos)
    val = np.array(val)
    sig = np.array(sig)
# Separate again relative maxima from relative minima
    max_val = val[sig > 0]
    max_pos = pos[sig > 0]
    min_val = val[sig < 0]
    min_pos = pos[sig < 0]
    del pos, val, sig, positions, signs, values
    return max_pos, max_val, min_pos, min_val


def min_max2D(data, half_width=3):
    """
    Find all relative minima and maxima in a 2D data matrix.

    A maximum is found if a value at position i of the vector is larger
    than or equal to all other values in a range [i-half_width:i+half_with]
    in both directions independently and at the same time strictly larger
    than all values of one side.

    The function searches first all relative extrema alng every column,
    then along every row. The returned positions are the combination of
    points found in both directions. In this way ridges parallel to the
    x axis and parallel to the y axes are detected.

    Parameters
    ----------
    data : 2D numpy float array
        Data matrix to be analysed
    half_width : int, optional (default: 3)
        Number of samples analyzed to all ides of every data sample.

    Returns
    -------
    maxima : List of two 1D numpy int arrays
        All position numbers where there is a relative maximum in array
        "data".
        maxima[0]: vector of row numbers, maxima[1]: vector of column
        numberss
    minima: List of two 1D numpy int arrays
        All position numbers where there is a relative minimum in array
        "data"
        minima[0]: vector of row numbers, minima[1]: vector of column
        numbers
    """
    ny, nx = data.shape
    extreme_pos = np.zeros((ny, nx), dtype=bool)
    extreme_neg = np.zeros((ny, nx), dtype=bool)
# Loop over columns
    for k in range(nx):
        max_pos, _, min_pos, _ = min_max(data[:, k], half_width=half_width)
        extreme_pos[max_pos, k] = True
        extreme_neg[min_pos, k] = True
# Loop over columns
    for k in range(ny):
        max_pos, _, min_pos, _ = min_max(data[k, :], half_width=half_width)
        extreme_pos[k, max_pos] = True
        extreme_neg[k, min_pos] = True
    return np.where(extreme_pos), np.where(extreme_neg)

# def zero_xing2D(self, data):
#     """
#     Search zero crossings in a 2D data set

#     The zero crossing is marked in the cell to the left or below the zero
#     crossing if the value itself is not zero.

#     Parameters
#     ----------
#     data : 2D numpy float array
#         data to be analyzed

#     Returns
#     -------
#     zeroes : List of two 1D numpy int arrays
#         All position numbers where there is a zero crossing in array
#         "data". zeroes[0]: vector of row numbers, zeroes[1]: vector of
#         column numbers

#     """
#     ny, nx = data.shape
#     xing = np.zeros((ny, nx), dtype=bool)
#     for k in range(nx-1):
#         for i in range(ny-1):
#             if data[i, k] == 0.:
#                 xing[i, k] = True
#                 continue
#             if data[i, k]*data[i+1, k] < 0. or\
#                     data[i, k]*data[i, k+1] < 0.:
#                 xing[i, k] = True
#     return np.where(xing)


def get_date_string():
    """
    Gets date and time of execution of the function

    Returns
    -------
    date and time: str
        Date and time are stored as YYYY-MM-DD_hh-mm-ss

    Is used to define output names where the text is standard but time of
    creation allows distinguishing.

    """
    now = datetime.now()
    c_time = now.strftime("%H-%M-%S")
    d = now.strftime("%Y-%m-%d")
    return f"{d}_{c_time}"


def file_name(base, ext):
    """
    Creates a filename composed of a header, the date string and an
    extension : base+"yyy-mm-dd_hh-mm-ss"+ext

    Parameters
    ----------
    base : str
        Header of file name
    ext : str
        Extension of file name. Must contain the dot. Utsually, the
        extension is simply a standard extension like ".dat", but it may
        contain more text (e.g., "corrected.dat")

    Returns
    -------
    file_name : str
        Composed file name

    """
    txt = get_date_string()
    return f"{base}_{txt}_{ext}"



# -*- coding: utf-8 -*-
# from .in_out import get_files, read_geography_file
# from .in_out import dialog, Geometrics
# from . import PyMaGra

# from .PyMaGra import Main
# print(f"Invoking __init__.py for {__name__}")



# -*- coding: utf-8 -*-
# Author:
#     Hermann Zeyen <hermann.zeyen@gmail.com>
#     Université Paris-Saclay, France
#
# License: BSD 3 clause

# from .io import get_files, read_geography_file, Data
# from .dialog import dialog
# from .geometrics import Geometrics

# print(f"Invoking __init__.py for {__name__}")
from .data import DataContainer
from .geometrics import Geometrics

__all__ = ["DataContainer", "Geometrics"]
